"""Test the synchronous convenience API intended for interactive use."""

# Note: before testing additional functionality (such as passing configs),
#       this should be migrated to mock-style testing rather than end-to-end

import lmstudio as lms

import pytest

from .support import (
    EXPECTED_EMBEDDING_ID,
    EXPECTED_LLM_ID,
    EXPECTED_VLM_ID,
    IMAGE_FILEPATH,
    TOOL_LLM_ID,
    closed_api_host,
)


@pytest.mark.lmstudio
def test_get_default_client() -> None:
    client = lms.get_default_client()
    assert isinstance(client, lms.Client)
    # Setting the API host after creation is disallowed (even if it is consistent)
    with pytest.raises(lms.LMStudioClientError, match="already created"):
        lms.get_default_client(client.api_host)
    # Ensure configured API host is used
    lms.sync_api._reset_default_client()
    try:
        with pytest.raises(lms.LMStudioClientError, match="not reachable"):
            lms.get_default_client(closed_api_host())
    finally:
        lms.sync_api._reset_default_client()


@pytest.mark.lmstudio
def test_configure_default_client() -> None:
    # Ensure the default client already exists
    client = lms.get_default_client()
    assert isinstance(client, lms.Client)
    # Setting the API host after creation is disallowed (even if it is consistent)
    with pytest.raises(lms.LMStudioClientError, match="already created"):
        lms.configure_default_client(client.api_host)
    # Ensure configured API host is used
    lms.sync_api._reset_default_client()
    try:
        lms.configure_default_client(closed_api_host())
        with pytest.raises(lms.LMStudioClientError, match="not reachable"):
            lms.get_default_client()
    finally:
        lms.sync_api._reset_default_client()


@pytest.mark.lmstudio
def test_llm_any() -> None:
    model = lms.llm()
    assert model.identifier in (EXPECTED_LLM_ID, EXPECTED_VLM_ID, TOOL_LLM_ID)


@pytest.mark.lmstudio
@pytest.mark.parametrize("model_id", (EXPECTED_LLM_ID, EXPECTED_VLM_ID, TOOL_LLM_ID))
def test_llm_specific(model_id: str) -> None:
    model = lms.llm(model_id)
    assert model.identifier == model_id


@pytest.mark.lmstudio
def test_embedding_any() -> None:
    model = lms.embedding_model()
    assert model.identifier == EXPECTED_EMBEDDING_ID


@pytest.mark.lmstudio
def test_embedding_specific() -> None:
    model = lms.embedding_model(EXPECTED_EMBEDDING_ID)
    assert model.identifier == EXPECTED_EMBEDDING_ID


@pytest.mark.lmstudio
def test_prepare_file() -> None:
    name = "example-file.txt"
    raw_data = b"raw data"
    file_handle = lms.sync_api._prepare_file(raw_data, name)
    assert file_handle.name == name
    assert file_handle.size_bytes == len(raw_data)
    assert file_handle.file_type == "text/plain"


@pytest.mark.lmstudio
def test_prepare_image() -> None:
    file_handle = lms.prepare_image(IMAGE_FILEPATH)
    assert file_handle.name == IMAGE_FILEPATH.name
    assert file_handle.size_bytes == len(IMAGE_FILEPATH.read_bytes())
    assert file_handle.file_type == "image"


@pytest.mark.lmstudio
def test_list_downloaded_models() -> None:
    all_models = [m.model_key for m in lms.list_downloaded_models()]
    embedding_models = [m.model_key for m in lms.list_downloaded_models("embedding")]
    llms = [m.model_key for m in lms.list_downloaded_models("llm")]
    assert set(all_models) == (set(embedding_models) | set(llms))


@pytest.mark.lmstudio
def test_list_loaded_models() -> None:
    all_models = [m.identifier for m in lms.list_loaded_models()]
    embedding_models = [m.identifier for m in lms.list_loaded_models("embedding")]
    llms = [m.identifier for m in lms.list_loaded_models("llm")]
    assert set(all_models) == (set(embedding_models) | set(llms))


@pytest.mark.lmstudio
def test_list_loaded_embedding_models() -> None:
    models = [m.identifier for m in lms.list_loaded_models("embedding")]
    assert not (set((EXPECTED_EMBEDDING_ID,)) - set(models))


@pytest.mark.lmstudio
def test_list_loaded_LLMs() -> None:
    models = [m.identifier for m in lms.list_loaded_models("llm")]
    assert not (set((EXPECTED_LLM_ID, EXPECTED_VLM_ID, TOOL_LLM_ID)) - set(models))



"""Test plugin examples can run as dev plugins."""

import subprocess
import sys
import time


from pathlib import Path
from queue import Empty, Queue
from threading import Thread
from typing import Iterable, TextIO

import pytest

from lmstudio.plugin._dev_runner import (
    _interrupt_child_process,
    _start_child_process,
    _PLUGIN_STOP_TIMEOUT,
)
from lmstudio.plugin.runner import _PLUGIN_API_STABILITY_WARNING


_THIS_DIR = Path(__file__).parent.resolve()
_PLUGIN_EXAMPLES_DIR = (_THIS_DIR / "../examples/plugins").resolve()


def _get_plugin_paths() -> list[Path]:
    return [p for p in _PLUGIN_EXAMPLES_DIR.iterdir() if p.is_dir()]


def _monitor_stream(stream: TextIO, queue: Queue[str], *, debug: bool = False) -> None:
    for line in stream:
        if debug:
            print(line)
        queue.put(line)


def _drain_queue(queue: Queue[str]) -> Iterable[str]:
    while True:
        try:
            yield queue.get(block=False)
        except Empty:
            break


def _exec_plugin(plugin_path: Path) -> subprocess.Popen[str]:
    # Run plugin in dev mode with IO pipes line buffered
    # (as the test process is monitoring for specific output)
    cmd = [
        sys.executable,
        "-u",
        "-m",
        "lmstudio.plugin",
        "--dev",
        str(plugin_path),
    ]
    return _start_child_process(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)


_PLUGIN_START_TIMEOUT = 5


def _exec_and_interrupt(plugin_path: Path) -> tuple[list[str], list[str], list[str]]:
    # Start the plugin in a child process
    process = _exec_plugin(plugin_path)
    # Ensure pipes don't fill up and block subprocess execution
    stdout_q: Queue[str] = Queue()
    stdout_thread = Thread(target=_monitor_stream, args=[process.stdout, stdout_q])
    stdout_thread.start()
    stderr_q: Queue[str] = Queue()
    stderr_thread = Thread(target=_monitor_stream, args=[process.stderr, stderr_q])
    stderr_thread.start()
    startup_lines: list[str] = []
    # Wait for plugin to start
    start_deadline = time.monotonic() + _PLUGIN_START_TIMEOUT
    try:
        print(f"Monitoring {stdout_q!r} for plugin started message")
        while True:
            remaining_time = start_deadline - time.monotonic()
            print(f"Waiting {remaining_time} seconds for plugin to start")
            try:
                line = stdout_q.get(timeout=remaining_time)
            except Empty:
                assert False, "Plugin subprocess failed to start"
            print(line)
            startup_lines.append(line)
            if "Ctrl-C to terminate" in line:
                break
    finally:
        # Instruct the process to terminate
        print("Sending termination request to plugin subprocess")
        stop_deadline = time.monotonic() + _PLUGIN_STOP_TIMEOUT
        _interrupt_child_process(process, (stop_deadline - time.monotonic()))
        # Give threads a chance to halt their file reads
        # (process terminating will close the pipes)
        stdout_thread.join(timeout=(stop_deadline - time.monotonic()))
        stderr_thread.join(timeout=(stop_deadline - time.monotonic()))
        with process:
            # Closes open pipes
            pass
    # Collect remainder of subprocess output
    shutdown_lines = [*_drain_queue(stdout_q)]
    stderr_lines = [*_drain_queue(stderr_q)]
    return startup_lines, shutdown_lines, stderr_lines


def _plugin_case_id(plugin_path: Path) -> str:
    return plugin_path.name


@pytest.mark.lmstudio
@pytest.mark.parametrize("plugin_path", _get_plugin_paths(), ids=_plugin_case_id)
def test_plugin_execution(plugin_path: Path) -> None:
    startup_lines, shutdown_lines, stderr_lines = _exec_and_interrupt(plugin_path)
    # Stderr should start with the API stability warning...
    warning_lines = [
        *_PLUGIN_API_STABILITY_WARNING.splitlines(keepends=True),
        "\n",
        "warnings.warn(_PLUGIN_API_STABILITY_WARNING, FutureWarning)\n",
    ]
    for warning_line in warning_lines:
        stderr_line = stderr_lines.pop(0)
        assert stderr_line.endswith(warning_line)
    # ... and then consist solely of logged information messages
    for log_line in stderr_lines:
        assert log_line.startswith("INFO:")
    # Startup should finish with the notification of how to terminate the dev plugin
    assert startup_lines[-1].endswith("Ctrl-C to terminate...\n")
    # Shutdown should finish with a graceful shutdown notice from the plugin runner
    assert shutdown_lines[-1] == "Plugin execution terminated by console interrupt\n"



"""Shared core async websocket implementation for the LM Studio remote access API."""

# Sync API: runs in dedicated background thread
# Async convenience API (once implemented): runs in dedicated background thread
# Async structured API: runs in foreground event loop

# Callback handling rules:
#
# * All callbacks are synchronous (use external async queues if needed)
# * All callbacks must be invoked from the *foreground* thread/event loop

import asyncio


# Python 3.10 compatibility: use concurrent.futures.TimeoutError instead of the builtin
# In 3.11+, these are the same type, in 3.10 futures have their own timeout exception
from concurrent.futures import Future as SyncFuture, TimeoutError as SyncFutureTimeout
from contextvars import ContextVar
from contextlib import AsyncExitStack, contextmanager
from functools import partial
from typing import (
    Any,
    Awaitable,
    Coroutine,
    Callable,
    ClassVar,
    Generator,
    TypeAlias,
    TypeVar,
)
from typing_extensions import (
    # Native in 3.11+
    Self,
)

from anyio import create_task_group, move_on_after
from httpx_ws import aconnect_ws, AsyncWebSocketSession, HTTPXWSException

from .schemas import DictObject
from .sdk_api import LMStudioRuntimeError
from .json_api import (
    LMStudioWebsocket,
    LMStudioWebsocketError,
    MultiplexingManager,
    RxQueue,
)

from ._logging import LogEventContext, new_logger

T = TypeVar("T")

__all__ = [
    "SyncFutureTimeout",
    "AsyncTaskManager",
    "AsyncWebsocketHandler",
]


class AsyncTaskManager:
    _LMS_TASK_MANAGER: ClassVar[ContextVar[Self]] = ContextVar("_LMS_TASK_MANAGER")

    def __init__(self, *, on_activation: Callable[[], Any] | None = None) -> None:
        self._activated = False
        self._event_loop: asyncio.AbstractEventLoop | None = None
        self._on_activation = on_activation
        self._task_queue: asyncio.Queue[Callable[[], Awaitable[Any]]] = asyncio.Queue()
        self._terminate = asyncio.Event()
        self._terminated = asyncio.Event()
        # For the case where the task manager is run via its context manager
        self._tm_started = asyncio.Event()
        self._tm_task: asyncio.Task[Any] | None = None

    ACTIVATION_TIMEOUT = 5  # Just starts an async task, should be fast
    TERMINATION_TIMEOUT = 20  # May have to shut down TCP links

    @property
    def activated(self) -> bool:
        return self._activated

    @property
    def active(self) -> bool:
        return (
            self._activated
            and self._event_loop is not None
            and not self._terminated.is_set()
        )

    async def __aenter__(self) -> Self:
        # Handle reentrancy the same way files do:
        # allow nested use as a CM, but close on the first exit
        if self._tm_task is None:
            self._tm_task = asyncio.create_task(self.run_until_terminated())
        with move_on_after(self.ACTIVATION_TIMEOUT):
            await self._tm_started.wait()
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.request_termination()
        with move_on_after(self.TERMINATION_TIMEOUT, shield=True):
            await self._terminated.wait()

    @classmethod
    def get_running_task_manager(cls) -> Self:
        try:
            return cls._LMS_TASK_MANAGER.get()
        except LookupError:
            err_msg = "No async task manager active in current context"
            raise LMStudioRuntimeError(err_msg) from None

    def ensure_running_in_task_loop(self) -> None:
        this_loop = self._event_loop
        if this_loop is None:
            # Task manager isn't active -> no coroutine can be running in it
            raise LMStudioRuntimeError(f"{self!r} is currently inactive.")
        try:
            running_loop = asyncio.get_running_loop()
        except RuntimeError:
            # No event loop in this thread -> can't be running in the task manager
            running_loop = None
        # Check if the running loop is the task manager's loop
        if running_loop is not this_loop:
            err_details = f"Expected: {this_loop!r} Running: {running_loop!r}"
            err_msg = f"{self!r} is running in a different event loop ({err_details})."
            raise LMStudioRuntimeError(err_msg)

    def is_running_in_task_loop(self) -> bool:
        try:
            self.ensure_running_in_task_loop()
        except LMStudioRuntimeError:
            return False
        return True

    def ensure_running_in_task_manager(self) -> None:
        # Task manager must be active in the running event loop
        self.ensure_running_in_task_loop()
        running_tm = self.get_running_task_manager()
        if running_tm is not self:
            err_details = f"Expected: {self!r} Running: {running_tm!r}"
            err_msg = f"Task is running in a different task manager ({err_details})."
            raise LMStudioRuntimeError(err_msg)

    async def request_termination(self) -> bool:
        """Request termination of the task manager from the same thread."""
        if not self.is_running_in_task_loop():
            return False
        if self._terminate.is_set():
            return False
        self._terminate.set()
        return True

    def request_termination_threadsafe(self) -> SyncFuture[bool]:
        """Request termination of the task manager from any thread."""
        loop = self._event_loop
        if loop is None:
            result: SyncFuture[bool] = SyncFuture()
            result.set_result(False)
            return result
        return self.run_coroutine_threadsafe(self.request_termination())

    async def wait_for_termination(self) -> None:
        """Wait in the same thread for the task manager to indicate it has terminated."""
        if not self.is_running_in_task_loop():
            return
        await self._terminated.wait()

    def wait_for_termination_threadsafe(self) -> None:
        """Wait in any thread for the task manager to indicate it has terminated."""
        loop = self._event_loop
        if loop is None:
            if not self._activated:
                raise RuntimeError(f"{self!r} is not yet active.")
            # Previously activated without an active event loop -> already terminated
            return
        self.run_coroutine_threadsafe(self.wait_for_termination()).result()

    async def terminate(self) -> None:
        """Terminate the task manager from the same thread."""
        if await self.request_termination():
            await self.wait_for_termination()

    def terminate_threadsafe(self) -> None:
        """Terminate the task manager from any thread."""
        if self.request_termination_threadsafe().result():
            self.wait_for_termination_threadsafe()

    def _mark_as_running(self: Self) -> None:
        # Explicit type hint to work around https://github.com/python/mypy/issues/16871
        if self._event_loop is not None:
            raise LMStudioRuntimeError("Async task manager is already running")
        self._event_loop = asyncio.get_running_loop()
        self._activated = True
        self._LMS_TASK_MANAGER.set(self)
        notify = self._on_activation
        if notify is not None:
            notify()
        self._tm_started.set()

    async def run_until_terminated(
        self, func: Callable[[], Coroutine[Any, Any, Any]] | None = None
    ) -> None:
        """Run task manager until termination is requested."""
        self._mark_as_running()
        # Use anyio and exceptiongroup to handle the lack of native task
        # and exception groups prior to Python 3.11
        try:
            async with create_task_group() as tg:
                tg.start_soon(self._accept_queued_tasks)
                if func is not None:
                    tg.start_soon(func)
                # Terminate all running tasks when termination is requested
                try:
                    await self._terminate.wait()
                finally:
                    tg.cancel_scope.cancel()
        finally:
            # Event loop is about to shut down
            self._terminated.set()
            self._event_loop = None

    async def _accept_queued_tasks(self) -> None:
        async with create_task_group() as additional_tasks:
            while True:
                task_func = await self._task_queue.get()
                additional_tasks.start_soon(task_func)

    async def schedule_task(self, func: Callable[[], Awaitable[Any]]) -> None:
        """Schedule given task in the task manager's base coroutine from the same thread.

        Important: task must NOT access any scoped resources from the scheduling scope.
        """
        self.ensure_running_in_task_loop()
        await self._task_queue.put(func)

    def schedule_task_threadsafe(self, func: Callable[[], Awaitable[Any]]) -> None:
        """Schedule given task in the task manager's base coroutine from any thread.

        Important: task must NOT access any scoped resources from the scheduling scope.
        """
        self.run_coroutine_threadsafe(self.schedule_task(func))

    def run_coroutine_threadsafe(self, coro: Coroutine[Any, Any, T]) -> SyncFuture[T]:
        """Call given coroutine in the task manager's event loop from any thread.

        Important: coroutine must NOT access any scoped resources from the calling scope.
        """
        loop = self._event_loop
        if loop is None:
            raise RuntimeError(f"{self!r} is currently inactive.")
        return asyncio.run_coroutine_threadsafe(coro, loop)

    def call_threadsafe(self, func: Callable[[], T]) -> SyncFuture[T]:
        """Call non-blocking function in the background event loop and make the result available.

        Important: function must NOT access any scoped resources from the calling scope.
        """

        async def coro() -> T:
            return func()

        return self.run_coroutine_threadsafe(coro())

    def call_soon_threadsafe(self, func: Callable[[], Any]) -> asyncio.Handle:
        """Call given non-blocking function in the background event loop."""
        loop = self._event_loop
        if loop is None:
            raise RuntimeError(f"{self!r} is currently inactive.")
        return loop.call_soon_threadsafe(func)


AsyncChannelInfo: TypeAlias = tuple[int, Callable[[], Awaitable[Any]]]
AsyncRemoteCallInfo: TypeAlias = tuple[int, Callable[[], Awaitable[Any]]]


class AsyncWebsocketHandler:
    """Async task handler for a single websocket connection."""

    WS_DISCONNECT_TIMEOUT = 10

    def __init__(
        self,
        task_manager: AsyncTaskManager,
        ws_url: str,
        auth_details: DictObject,
        log_context: LogEventContext | None = None,
    ) -> None:
        self._auth_details = auth_details
        self._connection_attempted = asyncio.Event()
        self._connection_failure: Exception | None = None
        self._auth_failure: Any | None = None
        self._task_manager = task_manager
        self._ws_url = ws_url
        self._ws: AsyncWebSocketSession | None = None
        self._ws_disconnected = asyncio.Event()
        self._rx_task: asyncio.Task[None] | None = None
        self._logger = logger = new_logger(type(self).__name__)
        logger.update_context(log_context, ws_url=ws_url)
        self._mux = MultiplexingManager(logger)

    async def connect(self) -> bool:
        """Connect websocket from the task manager's event loop."""
        task_manager = self._task_manager
        await task_manager.schedule_task(self._logged_ws_handler)
        await self._connection_attempted.wait()
        return self._ws is not None

    def connect_threadsafe(self) -> bool:
        """Connect websocket from any thread."""
        task_manager = self._task_manager
        task_manager.run_coroutine_threadsafe(self.connect()).result()
        return self._ws is not None

    async def disconnect(self) -> None:
        """Disconnect websocket from the task manager's event loop."""
        self._task_manager.ensure_running_in_task_loop()
        # Websocket handler task may already have been cancelled,
        # but the closure can be requested multiple times without issue
        self._ws_disconnected.set()
        ws = self._ws
        if ws is None:
            return
        await ws.close()

    def disconnect_threadsafe(self) -> None:
        """Disconnect websocket from any thread."""
        task_manager = self._task_manager
        task_manager.run_coroutine_threadsafe(self.disconnect()).result()

    async def _logged_ws_handler(self) -> None:
        self._logger.debug("Websocket handling task started")
        try:
            await self._handle_ws()
        except (asyncio.CancelledError, GeneratorExit):
            raise
        except BaseException:
            err_msg = "Terminating websocket task due to exception"
            self._logger.debug(err_msg, exc_info=True)
        finally:
            # Ensure connections attempt are unblocked even if the
            # background async task errors out completely
            self._connection_attempted.set()
            self._logger.debug("Websocket task terminated")

    async def _handle_ws(self) -> None:
        resources = AsyncExitStack()
        try:
            # For reliable shutdown, handler must run entirely inside the task manager
            self._task_manager.ensure_running_in_task_manager()
            ws: AsyncWebSocketSession = await resources.enter_async_context(
                aconnect_ws(self._ws_url)
            )
        except Exception as exc:
            self._connection_failure = exc
            raise

        def _clear_task_state() -> None:
            # Websocket is about to be disconnected (if it isn't already)
            self._ws = None

        resources.callback(_clear_task_state)
        async with resources:
            self._logger.debug("Websocket connected")
            self._ws = ws
            if not await self._authenticate():
                return
            self._connection_attempted.set()
            self._logger.info("Websocket session established")
            # Task will run until message reception fails or is cancelled
            try:
                await self._receive_messages()
            finally:
                self._logger.debug("Websocket demultiplexing task terminated.")
                # Notify foreground thread of background thread termination
                # (this covers termination due to link failure)
                await self.notify_client_termination()
                dc_timeout = self.WS_DISCONNECT_TIMEOUT
                with move_on_after(dc_timeout, shield=True) as cancel_scope:
                    # Workaround an anyio/httpx-ws issue with task cancellation:
                    # https://github.com/frankie567/httpx-ws/issues/107
                    self._ws = None
                    try:
                        await ws.close()
                    except Exception:
                        # Closing may fail if the link is already down
                        pass
                if cancel_scope.cancelled_caught:
                    self._logger.warn(
                        f"Failed to close websocket in {dc_timeout} seconds."
                    )
                else:
                    self._logger.info("Websocket closed.")

    async def send_json(self, message: DictObject) -> None:
        # This is only called if the websocket has been created
        self._task_manager.ensure_running_in_task_loop()
        ws = self._ws
        if ws is None:
            # Assume app is shutting down and the owning task has already been cancelled
            rx_queue = self._mux.map_tx_message(message)
            if rx_queue is not None:
                await rx_queue.put(None)
            return
        try:
            await ws.send_json(message)
        except Exception as exc:
            err = LMStudioWebsocket._get_tx_error(message, exc)
            # Log the underlying exception info, but simplify the raised traceback
            self._logger.debug(str(err), exc_info=True)
            raise err from None

    def send_json_threadsafe(self, message: DictObject) -> None:
        future = self._task_manager.run_coroutine_threadsafe(self.send_json(message))
        future.result()  # Block until the message is sent

    def run_background_coroutine(self, coro: Coroutine[Any, Any, T]) -> T:
        """Run given coroutine in the event loop and wait for the result."""
        return self._task_manager.run_coroutine_threadsafe(coro).result()

    @contextmanager
    def open_channel(self) -> Generator[AsyncChannelInfo, None, None]:
        self._task_manager.ensure_running_in_task_loop()
        rx_queue: RxQueue = asyncio.Queue()
        with self._mux.assign_channel_id(rx_queue) as call_id:
            yield call_id, rx_queue.get

    @contextmanager
    def start_call(self) -> Generator[AsyncRemoteCallInfo, None, None]:
        self._task_manager.ensure_running_in_task_loop()
        rx_queue: RxQueue = asyncio.Queue()
        with self._mux.assign_call_id(rx_queue) as call_id:
            yield call_id, rx_queue.get

    def new_threadsafe_rx_queue(self) -> tuple[RxQueue, Callable[[float | None], Any]]:
        rx_queue: RxQueue = asyncio.Queue()
        return rx_queue, partial(self._rx_queue_get_threadsafe, rx_queue)

    def acquire_channel_id_threadsafe(self, rx_queue: RxQueue) -> int:
        future = self._task_manager.call_threadsafe(
            partial(self._mux.acquire_channel_id, rx_queue)
        )
        return future.result()  # Wait for background thread to assign the ID

    def release_channel_id_threadsafe(self, channel_id: int, rx_queue: RxQueue) -> None:
        self._task_manager.call_soon_threadsafe(
            partial(self._mux.release_channel_id, channel_id, rx_queue)
        )

    def acquire_call_id_threadsafe(self, rx_queue: RxQueue) -> int:
        future = self._task_manager.call_threadsafe(
            partial(self._mux.acquire_call_id, rx_queue)
        )
        return future.result()  # Wait for background thread to assign the ID

    def release_call_id_threadsafe(self, call_id: int, rx_queue: RxQueue) -> None:
        self._task_manager.call_soon_threadsafe(
            partial(self._mux.release_call_id, call_id, rx_queue)
        )

    def _rx_queue_get_threadsafe(self, rx_queue: RxQueue, timeout: float | None) -> Any:
        future = self._task_manager.run_coroutine_threadsafe(rx_queue.get())
        try:
            return future.result(timeout)
        except SyncFutureTimeout:
            future.cancel()
            raise

    async def _receive_json(self) -> Any:
        # This is only called if the websocket has been created
        if __debug__:
            # This should only be called as part of the self._handle_ws task
            self._task_manager.ensure_running_in_task_manager()
        ws = self._ws
        if ws is None:
            # Assume app is shutting down and the owning task has already been cancelled
            return
        try:
            return await ws.receive_json()
        except Exception as exc:
            err = LMStudioWebsocket._get_rx_error(exc)
            # Log the underlying exception info, but simplify the raised traceback
            self._logger.debug(str(err), exc_info=True)
            raise err from None

    async def _authenticate(self) -> bool:
        # This is only called if the websocket has been created
        if __debug__:
            # This should only be called as part of the self._handle_ws task
            self._task_manager.ensure_running_in_task_manager()
        ws = self._ws
        if ws is None:
            # Assume app is shutting down and the owning task has already been cancelled
            return False
        auth_message = self._auth_details
        await self.send_json(auth_message)
        auth_result = await self._receive_json()
        self._logger.debug("Websocket authenticated", json=auth_result)
        if not auth_result["success"]:
            self._auth_failure = auth_result["error"]
            return False
        return True

    async def _process_next_message(self) -> bool:
        """Process the next message received on the websocket.

        Returns True if a message queue was updated.
        """
        # This is only called if the websocket has been created
        if __debug__:
            # This should only be called as part of the self._handle_ws task
            self._task_manager.ensure_running_in_task_manager()
        ws = self._ws
        if ws is None:
            # Assume app is shutting down and the owning task has already been cancelled
            return False
        message = await ws.receive_json()
        return await self._enqueue_message(message)

    async def _receive_messages(self) -> None:
        """Process received messages until task is cancelled."""
        while True:
            try:
                await self._process_next_message()
            except (LMStudioWebsocketError, HTTPXWSException):
                if self._ws is not None and not self._ws_disconnected.is_set():
                    # Websocket failed unexpectedly (rather than due to client shutdown)
                    self._logger.error("Websocket failed, terminating session.")
                break

    async def _enqueue_message(self, message: Any) -> bool:
        if message is None:
            self._logger.info(f"Websocket session failed ({self._ws_url})")
            self._ws = None
            return await self.notify_client_termination() > 0
        rx_queue = self._mux.map_rx_message(message)
        if rx_queue is None:
            return False
        await rx_queue.put(message)
        return True

    async def notify_client_termination(self) -> int:
        """Send None to all clients with open receive queues (from background thread)."""
        num_clients = 0
        for rx_queue in self._mux.all_queues():
            await rx_queue.put(None)
            num_clients += 1
        self._logger.debug(
            f"Notified {num_clients} clients of websocket termination",
            num_clients=num_clients,
        )
        return num_clients

    def notify_client_termination_threadsafe(self) -> int:
        """Send None to all clients with open receive queues (from foreground thread)."""
        return self.run_background_coroutine(self.notify_client_termination())



"""Structured JSON values for LM Studio APIs."""

import json

from typing import (
    Any,
    ClassVar,
    Generic,
    Mapping,
    MutableMapping,
    Protocol,
    Sequence,
    TypeAlias,
    TypedDict,
    TypeVar,
    cast,
    runtime_checkable,
)
from typing_extensions import (
    # Native in 3.11+
    Self,
)

from msgspec import Struct, ValidationError, convert, to_builtins
from msgspec.json import schema

from .sdk_api import LMStudioValueError, sdk_public_api, sdk_public_type

__all__ = [
    "AnyLMStudioStruct",
    "BaseModel",
    "DictObject",
    "DictSchema",
    "ModelSchema",
    "ValidationError",
]

DictObject: TypeAlias = Mapping[str, Any]  # Any JSON-compatible string-keyed dict
MutableDictObject: TypeAlias = MutableMapping[str, Any]
DictSchema: TypeAlias = Mapping[str, Any]  # JSON schema as a string-keyed dict
# It would be nice to require a "type" key in DictSchema, but that's currently tricky
# without "extra_items" support in TypedDict: https://peps.python.org/pep-0728/


def _format_json(data: Any, *, sort_keys: bool = True) -> str:
    return json.dumps(data, indent=2, sort_keys=sort_keys)


def _to_json_schema(cls: type, *, omit: Sequence[str] = ()) -> DictSchema:
    json_schema = schema(cls)
    schema_ref: str | None = json_schema.get("$ref", None)
    if schema_ref is not None:
        # The LM Studio API doesn't like schema definitions with
        # top level "$ref" keys referring to a named subschema,
        # so extract the target schema and use its field values
        # (this isn't fully general, but it works for `msgspec`)
        del json_schema["$ref"]
        _, schema_defs_key, schema_name = schema_ref.split("/")
        subschemas: dict[str, dict[str, Any]] = json_schema[schema_defs_key]
        named_schema = subschemas.pop(schema_name, None)
        if named_schema is None:
            raise LMStudioValueError(
                f"Could not resolve {schema_ref!r} in {json_schema!r}"
            )
        for field in omit:
            named_schema.pop(field, None)
        json_schema.update(named_schema)

    # msgspec automatically handles default values in the generated JSON schema
    # when they are properly defined in the Struct fields

    return json_schema


@sdk_public_type
@runtime_checkable
class ModelSchema(Protocol):
    """Protocol for classes that provide a JSON schema for their model."""

    @classmethod
    def model_json_schema(cls) -> DictSchema:
        """Return a JSON schema dict describing this model."""
        ...


@sdk_public_type
class BaseModel(Struct, omit_defaults=True, kw_only=True):
    """Base class for structured prediction output formatting."""

    # Allows structured predictions using a `pydantic.BaseModel` inspired format,
    # even in applications that don't otherwise depend on Pydantic

    @classmethod
    def model_json_schema(cls) -> DictSchema:
        """Returns JSON Schema dict describing the format of this class."""
        # Schema descriptions are always converted to dict[str, Any], but
        # Python type checkers don't know that. Use `cast` to notify them.
        return cast(
            dict[str, Any], to_builtins(_to_json_schema(cls), order="deterministic")
        )


_CAMEL_CASE_OVERRIDES = {
    # `_kv_` in snake_case becomes KV in camelCase
    "useFp16ForKvCache": "useFp16ForKVCache",
    "offloadKvCacheToGpu": "offloadKVCacheToGpu",
}

_SKIP_FIELD_RECURSION = set(
    (
        "json_schema",
        "jsonSchema",
    )
)


def _snake_case_to_camelCase(key: str) -> str:
    first, *rest = key.split("_")
    camelCase = "".join((first, *(w.capitalize() for w in rest)))
    return _CAMEL_CASE_OVERRIDES.get(camelCase, camelCase)


# TODO: Rework this conversion to be based on the API struct definitions
#       * Only recurse into fields that allow substructs
#       * Only check fields with a snake case -> camel case name conversion
def _snake_case_keys_to_camelCase(data: DictObject) -> DictObject:
    translated_data: dict[str, Any] = {}
    dicts_to_process = [(data, translated_data)]
    queued_dict_ids = set((id(data),))

    def _queue_dict(input_dict: DictObject, output_dict: dict[str, Any]) -> None:
        dict_id = id(input_dict)
        if dict_id in queued_dict_ids:
            raise LMStudioValueError("Data structure cycles are not supported.")
        queued_dict_ids.add(dict_id)
        dicts_to_process.append((input_dict, output_dict))

    for input_dict, output_dict in dicts_to_process:
        for k, v in input_dict.items():
            match v:
                case {}:
                    if k in _SKIP_FIELD_RECURSION:
                        new_value = v
                    else:
                        new_dict: dict[str, Any] = {}
                        _queue_dict(v, new_dict)
                        new_value = new_dict
                case [*_]:
                    new_list: list[Any] = []
                    for item in v:
                        new_item: Any
                        match item:
                            case {}:
                                new_dict_item: dict[str, Any] = {}
                                _queue_dict(item, new_dict_item)
                                new_item = new_dict_item
                            case [*_]:
                                raise LMStudioValueError(
                                    "Lists of lists are not supported."
                                )
                            case _:
                                new_item = item
                        new_list.append(new_item)
                    new_value = new_list
                case _:
                    new_value = v
            if "_" in k:
                new_k = _snake_case_to_camelCase(k)
            else:
                new_k = k
            output_dict[new_k] = new_value
    return translated_data


# Typed dict values conform to `object`, rather than to `Any`
# https://github.com/python/mypy/issues/8994
TWireFormat = TypeVar("TWireFormat", bound=Mapping[str, object])


class LMStudioStruct(Generic[TWireFormat], Struct, omit_defaults=True, kw_only=True):
    """Base class for LM Studio-specific structured JSON values."""

    # Inherited metaclass keyword arguments:
    #
    # * "None" fields should be omitted, not sent as "null"
    # * Allow non-default fields after default fields
    #

    # This is actually defined in msgspec.Struct,
    # but is missing from the published type stubs:
    # https://github.com/jcrist/msgspec/pull/813
    __struct_encode_fields__: ClassVar[tuple[str, ...]]

    @classmethod
    def _from_any_api_dict(cls, data: DictObject) -> Self:
        """Attempt to create an instance from a camelCase string-keyed dict."""
        return convert(data, cls)

    @classmethod
    def _from_api_dict(cls, data: TWireFormat) -> Self:
        """Attempt to create an instance from a camelCase string-keyed dict."""
        # SDK code should use `_from_any_api_dict` directly rather than external casts
        return cls._from_any_api_dict(data)

    @classmethod
    def _from_any_dict(cls, data: DictObject) -> Self:
        """Attempt to create an instance from a camelCase or snake_case string-keyed dict."""
        # `convert` won't report validation errors for optional fields, so we always check
        # for snake_case keys. SDK code should use `_from_api_dict` instead when the keys
        # are known to already be using camelCase multi-word formatting.
        return convert(_snake_case_keys_to_camelCase(data), cls)

    @classmethod
    @sdk_public_api()
    def from_dict(cls, data: TWireFormat) -> Self:
        """Attempt to create an instance from a camelCase or snake_case string-keyed dict."""
        # Statically type checked API for creating instances from mappings
        # SDK code should use `_from_any_dict` directly rather than external casts
        return cls._from_any_dict(data)

    def to_dict(self) -> TWireFormat:
        """Convert instance to a camelCase string-keyed dictionary."""
        # Struct instances are always converted to their corresponding dict,
        # but Python type checkers don't know that. Use `cast` to notify them.
        return cast(TWireFormat, to_builtins(self, order="deterministic"))

    def __str__(self) -> str:
        type_name = type(self).__name__
        formatted_data = _format_json(self.to_dict())
        return f"{type_name}.from_dict({formatted_data})"


AnyLMStudioStruct = LMStudioStruct[Any]


class EmptyStruct(LMStudioStruct["EmptyDict"]):
    """LM Studio struct with no defined fields."""

    pass


class EmptyDict(TypedDict):
    """Wire format with no defined fields."""

    pass



"""Background thread async websocket implementation for the LM Studio remote access API."""

# Sync API
# Async convenience API (once implemented)

import asyncio
import threading
import weakref

from contextlib import contextmanager
from typing import (
    Any,
    Coroutine,
    Callable,
    Generator,
    TypeAlias,
    TypeVar,
)

from httpx_ws import AsyncWebSocketSession

from .schemas import DictObject

from ._logging import new_logger, LogEventContext
from ._ws_impl import AsyncTaskManager, AsyncWebsocketHandler

# Allow the core client websocket management to be shared across all SDK interaction APIs
# See https://discuss.python.org/t/daemon-threads-and-background-task-termination/77604
# (Note: this implementation has the elements needed to run on *current* Python versions
# and omits the generalised features that the SDK doesn't need)
T = TypeVar("T")


class BackgroundThread(threading.Thread):
    """Background async event loop thread."""

    def __init__(
        self,
        task_target: Callable[[], Coroutine[Any, Any, Any]] | None = None,
        name: str | None = None,
    ) -> None:
        # Accepts the same args as `threading.Thread`, *except*:
        #   * a  `task_target` coroutine replaces the `target` function
        #   * No `daemon` option (always runs as a daemon)
        # Variant: accept `debug` and `loop_factory` options to forward to `asyncio.run`
        # Alternative: accept a `task_runner` callback, defaulting to `asyncio.run`
        self._task_target = task_target
        self._loop_started = loop_started = threading.Event()
        self._task_manager = AsyncTaskManager(on_activation=loop_started.set)
        # Annoyingly, we have to mark the background thread as a daemon thread to
        # prevent hanging at shutdown. Even checking `sys.is_finalizing()` is inadequate
        # https://discuss.python.org/t/should-sys-is-finalizing-report-interpreter-finalization-instead-of-runtime-finalization/76695
        # TODO: skip thread daemonization when running in a subinterpreter
        # (and also disable the convenience API in subinterpreters to avoid hanging on shutdown)
        super().__init__(name=name, daemon=True)
        weakref.finalize(self, self.terminate)

    @property
    def task_manager(self) -> AsyncTaskManager:
        return self._task_manager

    def start(self, wait_for_loop: bool = True) -> None:
        """Start background thread and (optionally) wait for the event loop to be ready."""
        super().start()
        if wait_for_loop:
            self.wait_for_loop()

    def run(self) -> None:
        """Run an async event loop in the background thread."""
        # Only public to override threading.Thread.run
        asyncio.run(self._task_manager.run_until_terminated(self._task_target))

    def wait_for_loop(self) -> asyncio.AbstractEventLoop | None:
        """Wait for the event loop to start from a synchronous foreground thread."""
        if self._task_manager._event_loop is None and not self._task_manager.activated:
            self._loop_started.wait()
        return self._task_manager._event_loop

    async def wait_for_loop_async(self) -> asyncio.AbstractEventLoop | None:
        """Wait for the event loop to start from an asynchronous foreground thread."""
        return await asyncio.to_thread(self.wait_for_loop)

    def terminate(self) -> bool:
        """Request termination of the event loop from a synchronous foreground thread."""
        return self._task_manager.request_termination_threadsafe().result()

    async def terminate_async(self) -> bool:
        """Request termination of the event loop from an asynchronous foreground thread."""
        return await asyncio.to_thread(self.terminate)

    def schedule_background_task(self, func: Callable[[], Any]) -> None:
        """Schedule given task in the event loop from a synchronous foreground thread."""
        self._task_manager.schedule_task_threadsafe(func)

    async def schedule_background_task_async(self, func: Callable[[], Any]) -> None:
        """Schedule given task in the event loop from an asynchronous foreground thread."""
        return await asyncio.to_thread(self.schedule_background_task, func)

    def run_background_coroutine(self, coro: Coroutine[Any, Any, T]) -> T:
        """Run given coroutine in the event loop and wait for the result."""
        return self._task_manager.run_coroutine_threadsafe(coro).result()

    async def run_background_coroutine_async(self, coro: Coroutine[Any, Any, T]) -> T:
        """Run given coroutine in the event loop and await the result."""
        return await asyncio.to_thread(self.run_background_coroutine, coro)

    def call_in_background(self, func: Callable[[], Any]) -> None:
        """Call given non-blocking function in the background event loop."""
        self._task_manager.call_soon_threadsafe(func)


# By default, the weakref finalization atexit hook is registered lazily.
# This can lead to shutdown sequencing issues if SDK users attempt to access
# client instances (such as the default sync client) from atexit hooks
# registered at import time (so they may end up running after the weakref
# finalization hook has already terminated background threads)
# Creating this finalizer here ensures the weakref finalization hook is
# registered at import time, and hence runs *after* any such hooks
# (assuming the lmstudio SDK is imported before the hooks are registered)
def _register_weakref_atexit_hook() -> None:
    class C:
        pass

    weakref.finalize(C(), int)


_register_weakref_atexit_hook()
del _register_weakref_atexit_hook


class AsyncWebsocketThread(BackgroundThread):
    def __init__(self, log_context: LogEventContext | None = None) -> None:
        super().__init__(task_target=self._log_thread_execution)
        self._logger = logger = new_logger(type(self).__name__)
        logger.update_context(log_context, thread_id=self.name)

    async def _log_thread_execution(self) -> None:
        self._logger.info("Websocket handling thread started")
        never_set = asyncio.Event()
        try:
            # Run the event loop until termination is requested
            await never_set.wait()
        except (asyncio.CancelledError, GeneratorExit):
            raise
        except BaseException:
            err_msg = "Terminating websocket thread due to exception"
            self._logger.debug(err_msg, exc_info=True)
        finally:
            self._logger.info("Websocket thread terminated")


SyncChannelInfo: TypeAlias = tuple[int, Callable[[float | None], Any]]
SyncRemoteCallInfo: TypeAlias = tuple[int, Callable[[float | None], Any]]


class SyncToAsyncWebsocketBridge:
    def __init__(
        self,
        ws_thread: AsyncWebsocketThread,
        ws_url: str,
        auth_details: DictObject,
        log_context: LogEventContext,
    ) -> None:
        self._ws_handler = AsyncWebsocketHandler(
            ws_thread.task_manager,
            ws_url,
            auth_details,
            log_context,
        )
        self._logger = logger = new_logger(type(self).__name__)
        logger.update_context(log_context)

    def connect(self) -> bool:
        return self._ws_handler.connect_threadsafe()

    def disconnect(self) -> None:
        self._ws_handler.disconnect_threadsafe()

    def send_json(self, message: DictObject) -> None:
        self._ws_handler.send_json_threadsafe(message)

    @contextmanager
    def open_channel(self) -> Generator[SyncChannelInfo, None, None]:
        ws_handler = self._ws_handler
        rx_queue, getter = ws_handler.new_threadsafe_rx_queue()
        channel_id = ws_handler.acquire_channel_id_threadsafe(rx_queue)
        try:
            yield channel_id, getter
        finally:
            ws_handler.release_channel_id_threadsafe(channel_id, rx_queue)

    @contextmanager
    def start_call(self) -> Generator[SyncRemoteCallInfo, None, None]:
        ws_handler = self._ws_handler
        rx_queue, getter = ws_handler.new_threadsafe_rx_queue()
        call_id = ws_handler.acquire_call_id_threadsafe(rx_queue)
        try:
            yield call_id, getter
        finally:
            ws_handler.release_call_id_threadsafe(call_id, rx_queue)

    def notify_client_termination_threadsafe(self) -> int:
        """Send None to all clients with open receive queues (from foreground thread)."""
        return self._ws_handler.notify_client_termination_threadsafe()

    # These attributes are currently accessed directly...
    @property
    def _ws(self) -> AsyncWebSocketSession | None:
        return self._ws_handler._ws

    @property
    def _connection_failure(self) -> Exception | None:
        return self._ws_handler._connection_failure

    @property
    def _auth_failure(self) -> Any | None:
        return self._ws_handler._auth_failure



#####################################################################
# This files has been automatically generated from:
#   ../async/test_model_catalog_async.py
#
# DO NOT EDIT THIS FILE! Edit the async test case listed above,
# and regenerate the synchronous test cases with async2sync.py
#####################################################################
"""Test listing, downloading, and loading available models."""

import logging
from contextlib import nullcontext

from contextlib import suppress

import pytest
from pytest import LogCaptureFixture as LogCap
from pytest_subtests import SubTests

from lmstudio import Client, LMStudioModelNotFoundError, LMStudioServerError
from lmstudio.json_api import DownloadedModelBase, ModelHandleBase

from ..support import (
    LLM_LOAD_CONFIG,
    EXPECTED_LLM,
    EXPECTED_LLM_ID,
    EXPECTED_EMBEDDING,
    EXPECTED_EMBEDDING_ID,
    EXPECTED_VLM_ID,
    SMALL_LLM_ID,
    TOOL_LLM_ID,
    check_sdk_error,
)


@pytest.mark.lmstudio
def test_list_downloaded_llm_sync(caplog: LogCap, subtests: SubTests) -> None:
    caplog.set_level(logging.DEBUG)
    subtests_started = subtests_passed = 0
    expected_model: str | None = EXPECTED_LLM
    # Model namespace is omitted so at least one test covers the default value
    with Client() as client:
        downloaded_models = client.llm.list_downloaded()
        assert downloaded_models
        for m in downloaded_models:
            subtests_started += 1
            with subtests.test("Check downloaded model", m=m):
                assert isinstance(m, DownloadedModelBase)
                # Check directly accessible details
                assert m.type == m.info.type
                assert m.path == m.info.path
                assert m.model_key == m.info.model_key
                # Check for expected model
                assert m.type == "llm"
                if expected_model is not None:
                    # Check if this is the expected model
                    if m.path.lower().startswith(expected_model):
                        expected_model = None
                subtests_passed += 1
        # The expected model should be present
        assert expected_model is None

    # Work around pytest-subtests not showing full output when subtests fail
    # https://github.com/pytest-dev/pytest-subtests/issues/76
    assert subtests_passed == subtests_started, "Fail due to failed subtest(s)"


@pytest.mark.lmstudio
def test_list_downloaded_embedding_sync(caplog: LogCap, subtests: SubTests) -> None:
    caplog.set_level(logging.DEBUG)
    subtests_started = subtests_passed = 0
    expected_model: str | None = EXPECTED_EMBEDDING
    with Client() as client:
        downloaded_models = client.embedding.list_downloaded()
        assert downloaded_models
        for m in downloaded_models:
            subtests_started += 1
            with subtests.test("Check downloaded model", m=m):
                assert isinstance(m, DownloadedModelBase)
                # Check directly accessible details
                assert m.type == m.info.type
                assert m.path == m.info.path
                assert m.model_key == m.info.model_key
                # Check for expected model
                assert m.type == "embedding"
                if expected_model is not None:
                    # Check if this is the expected model
                    if m.path.lower().startswith(expected_model):
                        expected_model = None
                subtests_passed += 1
        # The expected model should be present
        assert expected_model is None

    # Work around pytest-subtests not failing the test case when subtests fail
    # https://github.com/pytest-dev/pytest-subtests/issues/76
    assert subtests_passed == subtests_started, "Fail due to failed subtest(s)"


@pytest.mark.lmstudio
def test_list_downloaded_models_sync(caplog: LogCap, subtests: SubTests) -> None:
    caplog.set_level(logging.DEBUG)
    subtests_started = subtests_passed = 0
    expected_llm: str | None = EXPECTED_LLM
    expected_embedding: str | None = EXPECTED_EMBEDDING
    with Client() as client:
        downloaded_models = client.system.list_downloaded_models()
        assert downloaded_models
        for m in downloaded_models:
            subtests_started += 1
            with subtests.test("Check downloaded model", m=m):
                assert isinstance(m, DownloadedModelBase)
                # Check for expected models
                if m.type == "llm":
                    if expected_llm is not None:
                        # Check if this is the expected LLM
                        if m.path.lower().startswith(expected_llm):
                            expected_llm = None
                elif m.type == "embedding":
                    if expected_embedding is not None:
                        # Check if this is the expected embedding
                        if m.path.lower().startswith(expected_embedding):
                            expected_embedding = None
                subtests_passed += 1
        # The expected models should be present
        assert expected_llm is None
        assert expected_embedding is None


@pytest.mark.lmstudio
def test_list_loaded_llm_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        loaded_models = client.llm.list_loaded()
        assert loaded_models
        assert all(isinstance(m, ModelHandleBase) for m in loaded_models)
        models = [m.identifier for m in loaded_models]
        assert not (set((EXPECTED_LLM_ID, EXPECTED_VLM_ID, TOOL_LLM_ID)) - set(models))


@pytest.mark.lmstudio
def test_list_loaded_embedding_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        loaded_models = client.embedding.list_loaded()
        assert loaded_models
        assert all(isinstance(m, ModelHandleBase) for m in loaded_models)
        models = [m.identifier for m in loaded_models]
        assert not (set((EXPECTED_EMBEDDING_ID,)) - set(models))


DUPLICATE_MODEL_ERROR = "Model load error.*already exists"


@pytest.mark.slow
@pytest.mark.lmstudio
def test_load_duplicate_llm_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        llm = client.llm
        initially_loaded_models = sorted(llm.list_loaded(), key=str)
        with pytest.raises(LMStudioServerError, match=DUPLICATE_MODEL_ERROR):
            # Server will reject an explicitly duplicated model ID
            llm.load_new_instance(EXPECTED_LLM, EXPECTED_LLM_ID, config=LLM_LOAD_CONFIG)
        # Let the server assign a new instance identifier
        new_instance = llm.load_new_instance(EXPECTED_LLM, config=LLM_LOAD_CONFIG)
        assigned_model_id = new_instance.identifier
        with_model_duplicated = sorted(llm.list_loaded(), key=str)
        llm.unload(assigned_model_id)
        # Check behaviour now the duplicated model has been unloaded
        assert len(with_model_duplicated) == len(initially_loaded_models) + 1
        model_id_prefix, _, model_id_suffix = assigned_model_id.partition(":")
        assert model_id_prefix == EXPECTED_LLM_ID
        assert model_id_suffix.isascii(), assigned_model_id
        assert model_id_suffix.isdecimal(), assigned_model_id
        with_model_removed = sorted(llm.list_loaded(), key=str)
        assert with_model_removed == initially_loaded_models


@pytest.mark.slow
@pytest.mark.lmstudio
def test_load_duplicate_embedding_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        embedding = client.embedding
        initially_loaded_models = sorted(embedding.list_loaded(), key=str)
        with pytest.raises(LMStudioServerError, match=DUPLICATE_MODEL_ERROR):
            # Server will reject an explicitly duplicated model ID
            embedding.load_new_instance(EXPECTED_EMBEDDING, EXPECTED_EMBEDDING_ID)
        # Let the server assign a new instance identifier
        new_instance = embedding.load_new_instance(EXPECTED_EMBEDDING)
        assigned_model_id = new_instance.identifier
        with_model_duplicated = sorted(embedding.list_loaded(), key=str)
        embedding.unload(assigned_model_id)
        # Check behaviour now the duplicated model has been unloaded
        assert len(with_model_duplicated) == len(initially_loaded_models) + 1
        model_id_prefix, _, model_id_suffix = assigned_model_id.partition(":")
        assert model_id_prefix == EXPECTED_EMBEDDING_ID
        assert model_id_suffix.isascii(), assigned_model_id
        assert model_id_suffix.isdecimal(), assigned_model_id
        with_model_removed = sorted(embedding.list_loaded(), key=str)
        assert with_model_removed == initially_loaded_models


@pytest.mark.lmstudio
def test_get_model_llm_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        model = client.llm.model(EXPECTED_LLM_ID)
        assert model.identifier == EXPECTED_LLM_ID


@pytest.mark.lmstudio
def test_get_model_embedding_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        model = client.embedding.model(EXPECTED_EMBEDDING_ID)
        assert model.identifier == EXPECTED_EMBEDDING_ID


@pytest.mark.lmstudio
def test_get_any_model_llm_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        model = client.llm.model()
        assert model.identifier in (EXPECTED_LLM_ID, EXPECTED_VLM_ID, TOOL_LLM_ID)


@pytest.mark.lmstudio
def test_get_any_model_embedding_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        model = client.embedding.model()
        assert model.identifier == EXPECTED_EMBEDDING_ID


@pytest.mark.lmstudio
def test_invalid_unload_request_llm_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        llm = client.llm
        # This should error rather than timing out,
        # but avoid any risk of the client hanging...
        with nullcontext():
            with pytest.raises(LMStudioModelNotFoundError) as exc_info:
                llm.unload("No such model")
            check_sdk_error(exc_info, __file__)


@pytest.mark.lmstudio
def test_invalid_unload_request_embedding_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        # This should error rather than timing out,
        # but avoid any risk of the client hanging...
        with nullcontext():
            with pytest.raises(LMStudioModelNotFoundError) as exc_info:
                client.embedding.unload("No such model")
            check_sdk_error(exc_info, __file__)


@pytest.mark.lmstudio
def test_get_or_load_when_loaded_llm_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        model = client.llm.model(EXPECTED_LLM)
        assert model.identifier == EXPECTED_LLM_ID


@pytest.mark.lmstudio
def test_get_or_load_when_loaded_embedding_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        model = client.embedding.model(EXPECTED_EMBEDDING)
        assert model.identifier == EXPECTED_EMBEDDING_ID


@pytest.mark.slow
@pytest.mark.lmstudio
def test_get_or_load_when_unloaded_llm_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        llm = client.llm
        with suppress(LMStudioModelNotFoundError):
            llm.unload(EXPECTED_LLM_ID)
        model = llm.model(EXPECTED_LLM_ID, config=LLM_LOAD_CONFIG)
        assert model.identifier == EXPECTED_LLM_ID
        # LM Studio may default to JIT handling for models loaded with `getOrLoad`,
        # so ensure we restore a regular non-JIT instance with no TTL set
        model.unload()
        model = llm.load_new_instance(EXPECTED_LLM_ID, config=LLM_LOAD_CONFIG, ttl=None)
        assert model.identifier == EXPECTED_LLM_ID


@pytest.mark.slow
@pytest.mark.lmstudio
def test_get_or_load_when_unloaded_embedding_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        embedding = client.embedding
        with suppress(LMStudioModelNotFoundError):
            embedding.unload(EXPECTED_EMBEDDING_ID)
        model = embedding.model(EXPECTED_EMBEDDING_ID)
        assert model.identifier == EXPECTED_EMBEDDING_ID
        # LM Studio may default to JIT handling for models loaded with `getOrLoad`,
        # so ensure we restore a regular non-JIT instance with no TTL set
        model.unload()
        model = embedding.load_new_instance(EXPECTED_EMBEDDING_ID, ttl=None)
        assert model.identifier == EXPECTED_EMBEDDING_ID


@pytest.mark.slow
@pytest.mark.lmstudio
def test_jit_unloading_sync(caplog: LogCap) -> None:
    # For the time being, only test the embedding vs LLM cross-namespace
    # JIT unloading (since that ensures the info type mixing is handled).
    # Assuming LM Studio eventually switches to per-namespace JIT unloading,
    # this can be split into separate LLM and embedding test cases at that time.
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        # Unload the non-JIT instance of the embedding model
        with suppress(LMStudioModelNotFoundError):
            client.embedding.unload(EXPECTED_EMBEDDING_ID)
        # Load a JIT instance of the embedding model
        model1 = client.embedding.model(EXPECTED_EMBEDDING_ID, ttl=300)
        assert model1.identifier == EXPECTED_EMBEDDING_ID
        model1_info = model1.get_info()
        assert model1_info.identifier == model1.identifier
        # Load a JIT instance of the small testing LLM
        # This will unload the JIT instance of the testing embedding model
        model2 = client.llm.model(SMALL_LLM_ID, ttl=300)
        assert model2.identifier == SMALL_LLM_ID
        model2_info = model2.get_info()
        assert model2_info.identifier == model2.identifier
        # Attempting to query the now unloaded JIT embedding model will fail
        with pytest.raises(LMStudioModelNotFoundError):
            model1.get_info()
        # Restore things to the way other test cases expect them to be
        model2.unload()
        model = client.embedding.load_new_instance(EXPECTED_EMBEDDING_ID, ttl=None)
        assert model.identifier == EXPECTED_EMBEDDING_ID

    # Check for expected log messages
    jit_unload_event = "Unloading other JIT model"
    jit_unload_messages_debug: list[str] = []
    jit_unload_messages_info: list[str] = []
    jit_unload_messages = {
        logging.DEBUG: jit_unload_messages_debug,
        logging.INFO: jit_unload_messages_info,
    }
    for _logger_name, log_level, message in caplog.record_tuples:
        if jit_unload_event not in message:
            continue
        jit_unload_messages[log_level].append(message)

    assert len(jit_unload_messages_info) == 1
    assert len(jit_unload_messages_debug) == 1

    info_message = jit_unload_messages_info[0]
    debug_message = jit_unload_messages_debug[0]
    # Ensure info message omits model info, but includes config guidance
    unload_notice = f'"event": "{jit_unload_event}"'
    assert unload_notice in info_message
    loading_model_notice = f'"model_key": "{SMALL_LLM_ID}"'
    assert loading_model_notice in info_message
    unloaded_model_notice = f'"unloaded_model_key": "{EXPECTED_EMBEDDING_ID}"'
    assert unloaded_model_notice in info_message
    assert '"suggestion": ' in info_message
    assert "disable this behavior" in info_message
    assert '"unloaded_model": ' not in info_message
    # Ensure debug message includes model info, but omits config guidance
    assert unload_notice in debug_message
    assert loading_model_notice in info_message
    assert unloaded_model_notice in debug_message
    assert '"suggestion": ' not in debug_message
    assert "disable this behavior" not in debug_message
    assert '"unloaded_model": ' in debug_message



"""Plugin API client implementation."""

# Plugins are expected to maintain multiple concurrently open channels and handle
# multiple concurrent server requests, so plugin implementations are always async

import asyncio
import json
import os
import runpy
import sys
import warnings

from functools import partial
from pathlib import Path
from typing import Any, Awaitable, Callable, TypeAlias, TypeVar

from anyio import create_task_group

from .._logging import new_logger
from ..sdk_api import LMStudioFileNotFoundError, sdk_public_api
from ..schemas import DictObject
from ..async_api import AsyncClient
from .._sdk_models import (
    PluginsRpcSetConfigSchematicsParameter as SetConfigSchematicsParam,
    PluginsRpcSetGlobalConfigSchematicsParameter as SetGlobalConfigSchematicsParam,
)
from .sdk_api import LMStudioPluginInitError, LMStudioPluginRuntimeError
from .config_schemas import BaseConfigSchema
from .hooks import (
    _AsyncSessionPlugins,
    TPluginConfigSchema,
    TGlobalConfigSchema,
    run_prompt_preprocessor,
    run_token_generator,
    run_tools_provider,
)

# Available as lmstudio.plugin.*
__all__ = [
    "run_plugin",
    "run_plugin_async",
]

# Warn about the plugin API stability, since it is still experimental
_PLUGIN_API_STABILITY_WARNING = """\
Note the plugin API is not yet stable and may change without notice in future releases
"""

AnyHookImpl: TypeAlias = Callable[..., Awaitable[Any]]
THookImpl = TypeVar("THookImpl", bound=AnyHookImpl)
ReadyCallback: TypeAlias = Callable[[], Any]
HookRunner: TypeAlias = Callable[
    [
        str,  # Plugin name
        THookImpl,
        type[TPluginConfigSchema],
        type[TGlobalConfigSchema],
        _AsyncSessionPlugins,
        ReadyCallback,
    ],
    Awaitable[Any],
]

_HOOK_RUNNERS: dict[str, HookRunner[Any, Any, Any]] = {
    "preprocess_prompt": run_prompt_preprocessor,
    "generate_tokens": run_token_generator,
    "list_provided_tools": run_tools_provider,
}


class PluginClient(AsyncClient):
    def __init__(
        self,
        plugin_dir: str | os.PathLike[str],
        client_id: str | None = None,
        client_key: str | None = None,
    ) -> None:
        warnings.warn(_PLUGIN_API_STABILITY_WARNING, FutureWarning)
        self._client_id = client_id
        self._client_key = client_key
        super().__init__()
        # TODO: Consider moving file reading to class method and make this a data class
        self._plugin_path = plugin_path = Path(plugin_dir)
        manifest_path = plugin_path / "manifest.json"
        if not manifest_path.exists():
            raise LMStudioFileNotFoundError(manifest_path)
        manifest = json.loads(manifest_path.read_text(encoding="utf-8"))
        if manifest["type"] != "plugin":
            raise LMStudioPluginInitError(f"Invalid manifest type: {manifest['type']}")
        if manifest["runner"] != "python":
            # This works (even though the app doesn't natively support Python plugins yet),
            # as LM Studio doesn't check the runner type when requesting dev credentials.
            raise LMStudioPluginInitError(
                f'Invalid manifest runner: {manifest["runner"]} (expected "python")'
            )
        self.owner = manifest["owner"]
        self.name = name = manifest["name"]
        self._logger = logger = new_logger(__name__)
        logger.update_context(plugin=name)

    _ALL_SESSIONS = (
        # Plugin controller access all runs through a dedicated endpoint
        _AsyncSessionPlugins,
    )

    def _create_auth_message(self) -> DictObject:
        """Create an LM Studio websocket authentication message."""
        if self._client_id is None or self._client_key is None:
            return super()._create_auth_message()
        # Use plugin credentials to unlock the full plugin client API
        return self._format_auth_message(self._client_id, self._client_key)

    @property
    def plugins(self) -> _AsyncSessionPlugins:
        """Return the plugins API client session."""
        return self._get_session(_AsyncSessionPlugins)

    async def _run_hook_impl(
        self,
        hook_runner: HookRunner[THookImpl, TPluginConfigSchema, TGlobalConfigSchema],
        hook_impl: THookImpl,
        plugin_config_schema: type[TPluginConfigSchema],
        global_config_schema: type[TGlobalConfigSchema],
        notify_ready: ReadyCallback,
    ) -> None:
        """Run the given hook implementation."""
        await hook_runner(
            self.name,
            hook_impl,
            plugin_config_schema,
            global_config_schema,
            self.plugins,
            notify_ready,
        )

    _CONFIG_SCHEMA_SCOPES = {
        "plugin": ("ConfigSchema", "setConfigSchematics", SetConfigSchematicsParam),
        "global": (
            "GlobalConfigSchema",
            "setGlobalConfigSchematics",
            SetGlobalConfigSchematicsParam,
        ),
    }

    async def _load_config_schema(
        self, ns: DictObject, scope: str
    ) -> type[BaseConfigSchema]:
        logger = self._logger
        config_name, endpoint, param_type = self._CONFIG_SCHEMA_SCOPES[scope]
        maybe_config_schema = ns.get(config_name, None)
        if maybe_config_schema is None:
            # Use an empty config in the client, don't register any schema with the server
            logger.debug(f"Plugin does not define {config_name!r}")
            return BaseConfigSchema
        if not issubclass(maybe_config_schema, BaseConfigSchema):
            raise LMStudioPluginInitError(
                f"{config_name}: Expected {BaseConfigSchema!r} subclass definition, not {maybe_config_schema!r}"
            )
        config_schema: type[BaseConfigSchema] = maybe_config_schema
        kv_config_schematics = config_schema._to_kv_config_schematics()
        if kv_config_schematics is None:
            # No fields to configure, no need to register schema with the server
            logger.info(f"Plugin defines an empty {config_name!r}")
        else:
            # Only notify the server if there is at least one config field defined
            logger.info(f"Plugin defines {config_name!r}, sending to server...")
            await self.plugins.remote_call(
                endpoint,
                param_type(
                    schematics=kv_config_schematics,
                ),
            )
        return config_schema

    async def run_plugin(self, *, allow_local_imports: bool = False) -> int:
        # TODO: Nicer error handling
        plugin_path = self._plugin_path
        source_dir_path = plugin_path / "src"
        source_path = source_dir_path / "plugin.py"
        if not source_path.exists():
            raise LMStudioFileNotFoundError(source_path)
        # TODO: Consider passing this logger to hook runners (instead of each creating their own)
        logger = self._logger
        logger.update_context(plugin_name=self.name)
        logger.info(f"Running {source_path}")
        if allow_local_imports:
            # We don't try to revert the path change, as that can have odd side-effects
            sys.path.insert(0, str(source_dir_path))
        plugin_ns = runpy.run_path(str(source_path), run_name="__lms_plugin__")
        # Look up config schemas in the namespace
        plugin_schema = await self._load_config_schema(plugin_ns, "plugin")
        global_schema = await self._load_config_schema(plugin_ns, "global")
        # Look up hook implementations in the namespace
        implemented_hooks: list[Callable[[], Awaitable[Any]]] = []
        hook_ready_events: list[asyncio.Event] = []
        for hook_name, hook_runner in _HOOK_RUNNERS.items():
            hook_impl = plugin_ns.get(hook_name, None)
            if hook_impl is None:
                logger.debug(f"Plugin does not define the {hook_name!r} hook")
                continue
            logger.info(f"Plugin defines the {hook_name!r} hook")
            hook_ready_event = asyncio.Event()
            hook_ready_events.append(hook_ready_event)
            implemented_hooks.append(
                partial(
                    self._run_hook_impl,
                    hook_runner,
                    hook_impl,
                    plugin_schema,
                    global_schema,
                    hook_ready_event.set,
                )
            )
        plugin = self.name
        if not implemented_hooks:
            hook_list = "\n  - ".join(("", *sorted(_HOOK_RUNNERS)))
            print(
                f"No plugin hooks defined in {plugin!r}, "
                f"expected at least one of:{hook_list}"
            )
            return 1
        # Use anyio and exceptiongroup to handle the lack of native task
        # and exception groups prior to Python 3.11
        async with create_task_group() as tg:
            for implemented_hook in implemented_hooks:
                tg.start_soon(implemented_hook)
            # Should this have a time limit set to guard against SDK bugs?
            await asyncio.gather(*(e.wait() for e in hook_ready_events))
            await self.plugins.remote_call("pluginInitCompleted")
            # Indicate that prompt processing is ready
            print(
                f"Plugin {plugin!r} running, press Ctrl-C to terminate...", flush=True
            )
            # Task group will wait for the plugins to run
        return 0


ENV_CLIENT_ID = "LMS_PLUGIN_CLIENT_IDENTIFIER"
ENV_CLIENT_KEY = "LMS_PLUGIN_CLIENT_PASSKEY"


def get_plugin_credentials_from_env() -> tuple[str, str]:
    return os.environ[ENV_CLIENT_ID], os.environ[ENV_CLIENT_KEY]


@sdk_public_api()
async def run_plugin_async(
    plugin_dir: str | os.PathLike[str], *, allow_local_imports: bool = False
) -> None:
    """Asynchronously execute a plugin in development mode."""
    try:
        client_id, client_key = get_plugin_credentials_from_env()
    except KeyError:
        err_msg = f"ERROR: {ENV_CLIENT_ID} and {ENV_CLIENT_KEY} must both be set in the environment"
        raise LMStudioPluginRuntimeError(err_msg)
    async with PluginClient(plugin_dir, client_id, client_key) as plugin_client:
        await plugin_client.run_plugin(allow_local_imports=allow_local_imports)


@sdk_public_api()
def run_plugin(
    plugin_dir: str | os.PathLike[str], *, allow_local_imports: bool = False
) -> None:
    """Execute a plugin in application mode."""
    asyncio.run(run_plugin_async(plugin_dir, allow_local_imports=allow_local_imports))



"""Common definitions for defining the SDK API boundary."""

from contextlib import AsyncContextDecorator, ContextDecorator
from os import getenv
from types import TracebackType
from typing import Type, TypeVar
from typing_extensions import (
    # Native in 3.11+
    Self,
)

from ._logging import StructuredLogger

__all__ = [
    "LMStudioError",
    "LMStudioOSError",
    "LMStudioRuntimeError",
    "LMStudioValueError",
]

_SDK_MODULE = __name__.partition(".")[0]

_C = TypeVar("_C", bound=type)


def sdk_public_type(cls: _C) -> _C:
    """Indicates a class forms part of the public SDK boundary.

    Sets `__module__` to the top-level SDK import rather than
    leaving it set to the implementation module.

    Note: methods are *not* implicitly decorated as public SDK APIs
    """
    cls.__module__ = _SDK_MODULE
    return cls


@sdk_public_type
class LMStudioError(Exception):
    """Common base class for exceptions raised directly by the SDK.

    Note: exceptions raised by underlying libraries (such as websocket
          connection errors) are NOT wrapped as SDK exceptions.
    """


@sdk_public_type
class LMStudioOSError(OSError, LMStudioError):
    """The SDK received an error while accessing the local operating system."""


@sdk_public_type
class LMStudioFileNotFoundError(FileNotFoundError, LMStudioError):
    """The SDK failed to find the specified file on the local file system."""


@sdk_public_type
class LMStudioRuntimeError(RuntimeError, LMStudioError):
    """User requested an invalid sequence of operations from the SDK."""


@sdk_public_type
class LMStudioValueError(ValueError, LMStudioError):
    """User supplied an invalid value to the SDK."""


def _truncate_traceback(exc: BaseException | None) -> None:
    """Truncate API traceback at the SDK boundary (by default)."""
    if exc is None:
        return
    # Check env var every time so it can be updated at runtime
    if getenv("LMS_KEEP_INTERNAL_STACK"):
        # Setting this environment variable turns off the traceback filtering
        return
    if isinstance(exc, LMStudioError) or not isinstance(exc, Exception):
        # Truncate the traceback for SDK exceptions at the SDK boundary.
        # Also truncate asynchronous exceptions like KeyboardInterrupt.
        # Other unwrapped exceptions indicate SDK bugs and keep a full traceback.
        exc.__traceback__ = None


# We avoid contextlib.contextmanager when manipulating the way exceptions are
# reported, as it can have quirky impacts on whether changes actually take effect


class sdk_callback_invocation:
    """Catch and log raised exceptions to protect the message handling task."""

    def __init__(self, message: str, logger: StructuredLogger) -> None:
        self._message = message
        self._logger = logger

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> bool:
        if not isinstance(exc_val, Exception):
            return False
        # Callbacks are invoked from message handling tasks,
        # which need to keep going even if a callback fails
        # We don't want the SDK stack trace, but we do want
        # the callback exception details
        self._logger.error(self._message, exc_info=True, exc=repr(exc_val))
        return True


class sdk_public_api(ContextDecorator):
    """Indicates a callable forms part of the public SDK boundary.

    Also usable as a context manager (in both synchronous and
    asynchronous code) when publishing generator-based context
    managers and iterators.

    By default, tracebacks will be truncated at the SDK boundary.
    Set LMS_KEEP_INTERNAL_STACK to a non-empty string in the process
    environment to see full traceback information in SDK exceptions.
    """

    # ContextDecorator uses `functools.wraps`, so the original callable
    # can be accessed as `f.__wrapped__` when public APIs call each other
    # Unfortunately, MyPy complains if you do that:
    # https://github.com/python/typeshed/issues/13403

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        _truncate_traceback(exc_val)


class sdk_public_api_async(AsyncContextDecorator):
    """Indicates a coroutine forms part of the public SDK boundary.

    By default, tracebacks will be truncated at the SDK boundary.
    Set LMS_KEEP_INTERNAL_STACK to a non-empty string in the process
    environment to see full traceback information in SDK exceptions.
    """

    # While this *can* be used as an async context manager, there's
    # no compelling reason to ever use it that way (over just using
    # sdk_public_api as a synchronous context manager instead).

    # AsyncContextDecorator uses `functools.wraps`, so the original callable
    # can be accessed as `f.__wrapped__` when public APIs call each other
    # Unfortunately, MyPy complains if you do that:
    # https://github.com/python/typeshed/issues/13403

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        _truncate_traceback(exc_val)



#####################################################################
# This files has been automatically generated from:
#   ../async/test_images_async.py
#
# DO NOT EDIT THIS FILE! Edit the async test case listed above,
# and regenerate the synchronous test cases with async2sync.py
#####################################################################
"""Test uploading and predicting with vision models and images."""

import logging

import pytest
from pytest import LogCaptureFixture as LogCap

from io import BytesIO

from lmstudio import Client, Chat, FileHandle, LMStudioServerError, LocalFileInput

from ..support import (
    EXPECTED_VLM_ID,
    IMAGE_FILEPATH,
    SHORT_PREDICTION_CONFIG,
    VLM_PROMPT,
    check_sdk_error,
)

_IMAGE_DATA = IMAGE_FILEPATH.read_bytes()

_FILE_INPUT_CASES: list[tuple[str, LocalFileInput]] = [
    ("filesystem path", IMAGE_FILEPATH),
    ("bytes IO stream", BytesIO(_IMAGE_DATA)),
    ("raw bytes", _IMAGE_DATA),
    ("mutable buffer", bytearray(_IMAGE_DATA)),
    ("memoryview", memoryview(_IMAGE_DATA)),
]
_FILE_INPUT_CASE_IDS = [case[0] for case in _FILE_INPUT_CASES]


@pytest.mark.lmstudio
@pytest.mark.parametrize(
    "input_kind,file_input", _FILE_INPUT_CASES, ids=_FILE_INPUT_CASE_IDS
)
def test_prepare_sync(
    caplog: LogCap, input_kind: str, file_input: LocalFileInput
) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        session = client.files
        file = session._prepare_file(file_input)
        assert file
        assert isinstance(file, FileHandle)
        logging.info(f"Uploaded file from {input_kind}: {file}")
        image = session.prepare_image(file_input)
        assert image
        assert isinstance(image, FileHandle)
        logging.info(f"Uploaded image from {input_kind}: {image}")
        # Even with the same data uploaded, assigned identifiers should differ
        assert image != file


@pytest.mark.lmstudio
def test_prepare_from_file_obj_sync(caplog: LogCap) -> None:
    caplog.set_level(logging.DEBUG)
    with Client() as client:
        session = client.files
        with open(IMAGE_FILEPATH, "rb") as f:
            file = session._prepare_file(f)
        assert file
        assert isinstance(file, FileHandle)
        logging.info(f"Uploaded file from file object: {file}")
        with open(IMAGE_FILEPATH, "rb") as f:
            image = session.prepare_image(f)
        assert image
        assert isinstance(image, FileHandle)
        logging.info(f"Uploaded image from file object: {image}")
        # Even with the same data uploaded, assigned identifiers should differ
        assert image != file


@pytest.mark.slow
@pytest.mark.lmstudio
def test_vlm_predict_sync(caplog: LogCap) -> None:
    prompt = VLM_PROMPT
    caplog.set_level(logging.DEBUG)
    model_id = EXPECTED_VLM_ID
    with Client() as client:
        image_handle = client.files.prepare_image(IMAGE_FILEPATH)
        history = Chat()
        history.add_user_message((prompt, image_handle))
        vlm = client.llm.model(model_id)
        response = vlm.respond(history, config=SHORT_PREDICTION_CONFIG)
    logging.info(f"VLM response: {response!r}")
    assert response
    assert response.content
    assert isinstance(response.content, str)
    # Sometimes the VLM fails to call out the main color in the image
    assert "purple" in response.content or "image" in response.content


@pytest.mark.lmstudio
def test_non_vlm_predict_sync(caplog: LogCap) -> None:
    prompt = VLM_PROMPT
    caplog.set_level(logging.DEBUG)
    model_id = "hugging-quants/llama-3.2-1b-instruct"
    with Client() as client:
        image_handle = client.files.prepare_image(IMAGE_FILEPATH)
        history = Chat()
        history.add_user_message((prompt, image_handle))
        llm = client.llm.model(model_id)
        with pytest.raises(LMStudioServerError) as exc_info:
            llm.respond(history)
        check_sdk_error(exc_info, __file__)


@pytest.mark.slow
@pytest.mark.lmstudio
def test_vlm_predict_image_param_sync(caplog: LogCap) -> None:
    prompt = VLM_PROMPT
    caplog.set_level(logging.DEBUG)
    model_id = EXPECTED_VLM_ID
    with Client() as client:
        image_handle = client.files.prepare_image(IMAGE_FILEPATH)
        history = Chat()
        history.add_user_message(prompt, images=[image_handle])
        vlm = client.llm.model(model_id)
        response = vlm.respond(history, config=SHORT_PREDICTION_CONFIG)
    logging.info(f"VLM response: {response!r}")
    assert response
    assert response.content
    assert isinstance(response.content, str)
    # Sometimes the VLM fails to call out the main color in the image
    assert "purple" in response.content or "image" in response.content


@pytest.mark.lmstudio
def test_non_vlm_predict_image_param_sync(caplog: LogCap) -> None:
    prompt = VLM_PROMPT
    caplog.set_level(logging.DEBUG)
    model_id = "hugging-quants/llama-3.2-1b-instruct"
    with Client() as client:
        image_handle = client.files.prepare_image(IMAGE_FILEPATH)
        history = Chat()
        history.add_user_message(prompt, images=[image_handle])
        llm = client.llm.model(model_id)
        with pytest.raises(LMStudioServerError) as exc_info:
            llm.respond(history)
        check_sdk_error(exc_info, __file__)



"""Conversion between dicts/public config structs and server KVConfig(Stack)s."""

# Known KV config settings are defined in
# https://github.com/lmstudio-ai/lmstudio-js/blob/main/packages/lms-kv-config/src/schema.ts
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Container,
    Iterable,
    Sequence,
    Type,
    TypeAlias,
    TypeVar,
    cast,
    get_args,
)
from typing_extensions import (
    # Native in 3.11+
    assert_never,
)

from .sdk_api import LMStudioValueError
from .schemas import DictObject, DictSchema, ModelSchema, MutableDictObject
from ._sdk_models import (
    EmbeddingLoadModelConfig,
    EmbeddingLoadModelConfigDict,
    GpuSettingDict,
    GpuSplitConfig,
    GpuSplitConfigDict,
    KvConfig,
    KvConfigFieldDict,
    KvConfigStack,
    KvConfigStackLayerDict,
    LlmLoadModelConfig,
    LlmLoadModelConfigDict,
    LlmPredictionConfig,
    LlmPredictionConfigDict,
    LlmSplitStrategy,
    LlmStructuredPredictionSetting,
    LlmStructuredPredictionSettingDict,
    SerializedKVConfigSettings,
)


@dataclass(frozen=True)
class ConfigField:
    client_key: str

    def to_kv_field(
        self, server_key: str, client_config: DictObject
    ) -> KvConfigFieldDict | None:
        return {
            "key": server_key,
            "value": client_config[self.client_key],
        }

    def update_client_config(
        self, client_config: MutableDictObject, value: Any
    ) -> None:
        client_config[self.client_key] = value


@dataclass(frozen=True)
class CheckboxField(ConfigField):
    def to_kv_field(
        self, server_key: str, client_config: DictObject
    ) -> KvConfigFieldDict | None:
        return {
            "key": server_key,
            "value": {"checked": True, "value": client_config[self.client_key]},
        }

    def update_client_config(
        self, client_config: MutableDictObject, value: DictObject
    ) -> None:
        if value.get("checked", False):
            # Tolerate 'value' being omitted
            client_config[self.client_key] = value.get("value", None)


@dataclass(frozen=True)
class NestedKeyField(ConfigField):
    nested_key: str

    def to_kv_field(
        self, server_key: str, client_config: DictObject
    ) -> KvConfigFieldDict | None:
        containing_value = client_config[self.client_key]
        nested_key = self.nested_key
        if nested_key not in containing_value:
            return None
        return {
            "key": server_key,
            "value": containing_value[nested_key],
        }

    def update_client_config(
        self, client_config: MutableDictObject, value: Any
    ) -> None:
        containing_value = client_config.setdefault(self.client_key, {})
        containing_value[self.nested_key] = value


@dataclass(frozen=True)
class MultiPartField(ConfigField):
    nested_keys: tuple[str, ...]
    client_to_server: Callable[..., Any]
    server_to_client: Callable[[DictObject, MutableDictObject], None]

    def to_kv_field(
        self, server_key: str, client_config: DictObject
    ) -> KvConfigFieldDict | None:
        client_container: DictObject = client_config[self.client_key]
        values = (client_container.get(key, None) for key in self.nested_keys)
        return {
            "key": server_key,
            "value": self.client_to_server(*values),
        }

    def update_client_config(
        self, client_config: MutableDictObject, server_value: DictObject
    ) -> None:
        client_container: MutableDictObject = client_config.setdefault(
            self.client_key, {}
        )
        self.server_to_client(server_value, client_container)


# TODO: figure out a way to compare this module against the lmstudio-js mappings
# TODO: Define a JSON or TOML data file for mapping prediction config
#       fields to config stack entries (preferably JSON exported by
#       lmstudio-js rather than something maintained in the Python SDK)
#       https://github.com/lmstudio-ai/lmstudio-js/issues/253
_COMMON_LLAMA_LOAD_KEYS: DictObject = {
    "keepModelInMemory": ConfigField("keepModelInMemory"),
    "ropeFrequencyBase": CheckboxField("ropeFrequencyBase"),
    "ropeFrequencyScale": CheckboxField("ropeFrequencyScale"),
    "tryMmap": ConfigField("tryMmap"),
    "acceleration": {
        "offloadRatio": NestedKeyField("gpu", "ratio"),
    },
}

_COMMON_MODEL_LOAD_KEYS: DictObject = {
    "contextLength": ConfigField("contextLength"),
}


def _gpu_settings_to_gpu_split_config(
    main_gpu: int | None,
    llm_split_strategy: LlmSplitStrategy | None,
    disabledGpus: Sequence[int] | None,
) -> GpuSplitConfigDict:
    gpu_split_config: GpuSplitConfigDict = {
        "disabledGpus": [*disabledGpus] if disabledGpus else [],
        "strategy": "evenly",
        "priority": [],
        "customRatio": [],
    }
    match llm_split_strategy:
        case "evenly" | None:
            pass
        case "favorMainGpu":
            gpu_split_config["strategy"] = "priorityOrder"
            if main_gpu is not None:
                gpu_split_config["priority"] = [main_gpu]
        case _:
            if TYPE_CHECKING:
                assert_never(llm_split_strategy)
            err_msg = f"Unknown LLM GPU offload split strategy: {llm_split_strategy}"
            hint = f"Known strategies: {get_args(LlmSplitStrategy)}"
            raise LMStudioValueError(f"{err_msg} ({hint})")
    return gpu_split_config


def _gpu_split_config_to_gpu_settings(
    server_dict: DictObject, client_dict: MutableDictObject
) -> None:
    gpu_settings_dict: GpuSettingDict = cast(GpuSettingDict, client_dict)
    gpu_split_config = GpuSplitConfig._from_any_api_dict(server_dict)
    disabled_gpus = gpu_split_config.disabled_gpus
    if disabled_gpus is not None:
        gpu_settings_dict["disabledGpus"] = disabled_gpus
    match gpu_split_config.strategy:
        case "evenly":
            gpu_settings_dict["splitStrategy"] = "evenly"
        case "priorityOrder":
            # For now, this can only map to "favorMainGpu"
            # Skip reporting the GPU offload details otherwise
            priority = gpu_split_config.priority
            if priority is not None and len(priority) == 1:
                gpu_settings_dict["splitStrategy"] = "favorMainGpu"
                gpu_settings_dict["mainGpu"] = priority[0]
        case "custom":
            # Currently no way to set up or report custom offload settings
            pass
        case _:
            if TYPE_CHECKING:
                assert_never(gpu_split_config.strategy)
            # Simply don't report details for unknown server strategies


SUPPORTED_SERVER_KEYS: dict[str, DictObject] = {
    "load": {
        "gpuSplitConfig": MultiPartField(
            "gpu",
            ("mainGpu", "splitStrategy", "disabledGpus"),
            _gpu_settings_to_gpu_split_config,
            _gpu_split_config_to_gpu_settings,
        ),
        "gpuStrictVramCap": ConfigField("gpuStrictVramCap"),
    },
    "embedding.load": {
        **_COMMON_MODEL_LOAD_KEYS,
        "llama": _COMMON_LLAMA_LOAD_KEYS,
    },
    "llm.load": {
        **_COMMON_MODEL_LOAD_KEYS,
        "numExperts": ConfigField("numExperts"),
        "seed": CheckboxField("seed"),
        "offloadKVCacheToGpu": ConfigField("offloadKVCacheToGpu"),
        "llama": {
            **_COMMON_LLAMA_LOAD_KEYS,
            "evalBatchSize": ConfigField("evalBatchSize"),
            "flashAttention": ConfigField("flashAttention"),
            "kCacheQuantizationType": CheckboxField("llamaKCacheQuantizationType"),
            "vCacheQuantizationType": CheckboxField("llamaVCacheQuantizationType"),
            "useFp16ForKVCache": ConfigField("useFp16ForKVCache"),
        },
    },
    "llm.prediction": {
        "contextOverflowPolicy": ConfigField("contextOverflowPolicy"),
        "maxPredictedTokens": CheckboxField("maxTokens"),  # Shorter name in client API
        "minPSampling": CheckboxField("minPSampling"),
        "promptTemplate": ConfigField("promptTemplate"),
        "repeatPenalty": CheckboxField("repeatPenalty"),
        "stopStrings": ConfigField("stopStrings"),
        "structured": ConfigField("structured"),
        "temperature": ConfigField("temperature"),
        "toolCallStopStrings": ConfigField("toolCallStopStrings"),
        "tools": ConfigField("rawTools"),  # Encourage calling .act() instead
        "topKSampling": ConfigField("topKSampling"),
        "topPSampling": CheckboxField("topPSampling"),
        "llama": {
            # Nested KV structure is flattened in client API
            "cpuThreads": ConfigField("cpuThreads"),
        },
        "reasoning": {
            # Nested KV structure is flattened in client API
            "parsing": ConfigField("reasoningParsing"),
        },
        "speculativeDecoding": {
            # Nested KV structure is flattened in client API
            "draftModel": ConfigField("draftModel"),
            "minDraftLengthToConsider": ConfigField(
                "speculativeDecodingMinDraftLengthToConsider"
            ),
            "minContinueDraftingProbability": ConfigField(
                "speculativeDecodingMinContinueDraftingProbability"
            ),
            "numDraftTokensExact": ConfigField(
                "speculativeDecodingNumDraftTokensExact"
            ),
        },
    },
}


# Define mappings to translate server KV configs to client config instances
def _iter_server_keys(
    *namespaces: str, excluded: Container[str] = ()
) -> Iterable[tuple[str, ConfigField]]:
    # Map dotted config field names to their client config field counterparts
    for namespace in namespaces:
        scopes: list[tuple[str, DictObject]] = [
            (namespace, SUPPORTED_SERVER_KEYS[namespace])
        ]
        for prefix, scope in scopes:
            for k, v in scope.items():
                if k in excluded:
                    # 'load' config namespace currently includes some LLM-only config keys
                    continue
                prefixed_key = f"{prefix}.{k}" if prefix else k
                if isinstance(v, ConfigField):
                    yield prefixed_key, v
                else:
                    assert isinstance(v, dict)
                    scopes.append((prefixed_key, v))


FROM_SERVER_LOAD_LLM = dict(_iter_server_keys("load", "llm.load"))
FROM_SERVER_LOAD_EMBEDDING = dict(
    _iter_server_keys("load", "embedding.load", excluded="gpuStrictVramCap")
)
FROM_SERVER_PREDICTION = dict(_iter_server_keys("llm.prediction"))
FROM_SERVER_CONFIG = dict(_iter_server_keys(*SUPPORTED_SERVER_KEYS))


# Define mappings to translate client config instances to server KV configs
FromServerKeymap = dict[str, ConfigField]
ToServerKeymap = dict[str, list[tuple[str, ConfigField]]]


def _invert_config_keymap(from_server: FromServerKeymap) -> ToServerKeymap:
    to_server: ToServerKeymap = {}
    for server_key, config_field in sorted(from_server.items()):
        client_key = config_field.client_key
        # There's at least one client field (gpu) which maps to
        # multiple KV config fields, so don't expect a 1:1 mapping
        config_fields = to_server.setdefault(client_key, [])
        config_fields.append((server_key, config_field))
    return to_server


TO_SERVER_LOAD_LLM = _invert_config_keymap(FROM_SERVER_LOAD_LLM)
TO_SERVER_LOAD_EMBEDDING = _invert_config_keymap(FROM_SERVER_LOAD_EMBEDDING)
TO_SERVER_PREDICTION = _invert_config_keymap(FROM_SERVER_PREDICTION)


TLoadConfig = TypeVar("TLoadConfig", LlmLoadModelConfig, EmbeddingLoadModelConfig)
TLoadConfigDict = TypeVar(
    "TLoadConfigDict", LlmLoadModelConfigDict, EmbeddingLoadModelConfigDict
)


def dict_from_kvconfig(config: KvConfig | SerializedKVConfigSettings) -> DictObject:
    return {kv.key: kv.value for kv in config.fields}


def parse_server_config(server_config: DictObject) -> DictObject:
    """Map server config fields to client config fields."""
    result: MutableDictObject = {}
    for kv in server_config.get("fields", []):
        key = kv["key"]
        config_field = FROM_SERVER_CONFIG.get(key, None)
        if config_field is None:
            # Skip unknown keys (server might be newer than the SDK)
            continue
        value = kv.get("value", None)  # Tolerate 'value' being omitted
        config_field.update_client_config(result, value)
    return result


def parse_llm_load_config(server_config: DictObject) -> LlmLoadModelConfig:
    return LlmLoadModelConfig._from_any_api_dict(parse_server_config(server_config))


def parse_prediction_config(server_config: DictObject) -> LlmPredictionConfig:
    return LlmPredictionConfig._from_any_api_dict(parse_server_config(server_config))


def _api_override_kv_config_stack(
    fields: list[KvConfigFieldDict],
    additional_layers: Sequence[KvConfigStackLayerDict] = (),
) -> KvConfigStack:
    return KvConfigStack._from_api_dict(
        {
            "layers": [
                {
                    "layerName": "apiOverride",
                    "config": {
                        "fields": fields,
                    },
                },
                *additional_layers,
            ],
        }
    )


def _to_kv_config_stack_base(
    config: DictObject, keymap: ToServerKeymap
) -> list[KvConfigFieldDict]:
    fields: list[KvConfigFieldDict] = []
    remaining_keys = set(config.keys())
    for client_key, config_fields in keymap.items():
        if client_key not in config:
            continue
        remaining_keys.remove(client_key)
        for server_key, config_field in config_fields:
            kv_field = config_field.to_kv_field(server_key, config)
            if kv_field is not None:
                fields.append(kv_field)
    if remaining_keys:
        raise LMStudioValueError(f"Unknown config settings: {sorted(remaining_keys)}")
    return fields


def _client_config_to_kv_config_stack(
    config: DictObject, keymap: ToServerKeymap
) -> KvConfigStack:
    fields = _to_kv_config_stack_base(config, keymap)
    return _api_override_kv_config_stack(fields)


def load_config_to_kv_config_stack(
    config: TLoadConfig | DictObject | None, config_type: Type[TLoadConfig]
) -> KvConfigStack:
    """Helper to convert load configs to KvConfigStack instances with strict typing."""
    dict_config: DictObject
    if config is None:
        dict_config = {}
    elif isinstance(config, config_type):
        dict_config = config.to_dict()
    else:
        assert isinstance(config, dict)
        dict_config = config_type._from_any_dict(config).to_dict()
    if config_type is LlmLoadModelConfig:
        return _client_config_to_kv_config_stack(dict_config, TO_SERVER_LOAD_LLM)
    assert config_type is EmbeddingLoadModelConfig
    return _client_config_to_kv_config_stack(dict_config, TO_SERVER_LOAD_EMBEDDING)


ResponseSchema: TypeAlias = (
    DictSchema
    | LlmStructuredPredictionSetting
    | LlmStructuredPredictionSettingDict
    | type[ModelSchema]
)


def prediction_config_to_kv_config_stack(
    response_format: Type[ModelSchema] | ResponseSchema | None,
    config: LlmPredictionConfig | LlmPredictionConfigDict | None,
    for_text_completion: bool = False,
) -> tuple[bool, KvConfigStack]:
    dict_config: LlmPredictionConfigDict
    if config is None:
        dict_config = {}
    elif isinstance(config, LlmPredictionConfig):
        dict_config = config.to_dict()
    else:
        assert isinstance(config, dict)
        dict_config = LlmPredictionConfig._from_any_dict(config).to_dict()
    if response_format is not None:
        if "structured" in dict_config:
            raise LMStudioValueError(
                "Cannot specify both 'response_format' in API call and 'structured' in config"
            )
        response_schema: LlmStructuredPredictionSettingDict
        structured = True
        if isinstance(response_format, LlmStructuredPredictionSetting):
            response_schema = response_format.to_dict()
        elif isinstance(response_format, type) and issubclass(
            response_format, ModelSchema
        ):
            response_schema = {
                "type": "json",
                "jsonSchema": response_format.model_json_schema(),
            }
        else:
            # Casts are needed as mypy doesn't detect that the given case patterns
            # conform to the definition of LlmStructuredPredictionSettingDict
            match response_format:
                case {"type": "json", "jsonSchema": _} as json_schema:
                    response_schema = cast(
                        LlmStructuredPredictionSettingDict, json_schema
                    )
                case {"type": "gbnf", "gbnfGrammar": _} as gbnf_schema:
                    response_schema = cast(
                        LlmStructuredPredictionSettingDict, gbnf_schema
                    )
                case {"type": _}:
                    # Assume any other input with a type key is a JSON schema definition
                    response_schema = {
                        "type": "json",
                        "jsonSchema": response_format,
                    }
                case _:
                    raise LMStudioValueError(
                        f"Failed to parse response format: {response_format!r}"
                    )
        dict_config["structured"] = response_schema
    else:
        # The response schema may also be passed in via the config
        # (doing it this way type hints as an unstructured result,
        # but we still allow it at runtime for consistency with JS)
        match dict_config:
            case {"structured": {"type": "json" | "gbnf"}}:
                structured = True
            case _:
                structured = False
    fields = _to_kv_config_stack_base(dict_config, TO_SERVER_PREDICTION)
    additional_layers: list[KvConfigStackLayerDict] = []
    if for_text_completion:
        additional_layers.append(_get_completion_config_layer())
    return structured, _api_override_kv_config_stack(fields, additional_layers)


def _get_completion_config_layer() -> KvConfigStackLayerDict:
    """Config layer to request text completion instead of a chat response."""
    # There is only one prediction endpoint in the LM Studio API, and it defaults to chat responses
    jinja_template = "{% for message in messages %}{{ message['content'] }}{% endfor %}"
    jinja_config = {
        "messagesConfig": {
            "contentConfig": {
                "type": "string",
            },
        },
        "useTools": False,
    }
    return {
        "layerName": "completeModeFormatting",
        "config": {
            "fields": [
                {
                    "key": "llm.prediction.promptTemplate",
                    "value": {
                        "type": "jinja",
                        "jinjaPromptTemplate": {
                            "bosToken": "",
                            "eosToken": "",
                            "template": jinja_template,
                            "inputConfig": jinja_config,
                        },
                        "stopStrings": [],
                    },
                }
            ],
        },
    }



"""Define plugin config schemas."""

from dataclasses import Field, dataclass, fields
from typing import (
    Any,
    ClassVar,
    Generic,
    Sequence,
    TypeVar,
    cast,
)

from typing_extensions import (
    # Native in 3.11+
    Self,
    dataclass_transform,
)

from .._kv_config import dict_from_kvconfig
from .._sdk_models import (
    SerializedKVConfigSchematics,
    SerializedKVConfigSchematicsField,
    SerializedKVConfigSettings,
)
from ..sdk_api import sdk_public_api
from .sdk_api import LMStudioPluginInitError, plugin_sdk_type

# Available as lmstudio.plugin.*
__all__ = [
    "BaseConfigSchema",
    "config_field",
]

_T = TypeVar("_T")

_CONFIG_FIELDS_KEY = "__kv_config_fields__"


@dataclass(frozen=True, slots=True)
class _ConfigField(Generic[_T]):
    """Plugin config field specification."""

    label: str
    hint: str

    @property
    def default(self) -> _T:
        """The default value for this config field."""
        # Defaults must be static values, as the UI isn't directly running any plugin code
        raise NotImplementedError

    # This never actually gets called (as __set_name__ switches to the default value at runtime)
    # However, it's here so type checkers accept field definitions as matching their value type
    def __get__(
        self, _obj: "BaseConfigSchema | None", _obj_type: type["BaseConfigSchema"]
    ) -> _T:
        return self.default

    def __set_name__(self, obj_type: type["BaseConfigSchema"], name: str) -> None:
        if not issubclass(obj_type, BaseConfigSchema):
            msg = f"Plugin config fields must be defined on {BaseConfigSchema.__name__} instances"
            raise LMStudioPluginInitError(msg)
        # Append this field to the config fields for this schema, creating the list if necessary
        config_fields: list[SerializedKVConfigSchematicsField]
        config_fields = obj_type.__dict__.get(_CONFIG_FIELDS_KEY, None)
        if config_fields is None:
            # First config field defined for this schema, so create the config field list
            config_fields = []
            try:
                inherited_fields = getattr(obj_type, _CONFIG_FIELDS_KEY)
            except AttributeError:
                pass
            else:
                # Any inherited fields are included first
                config_fields.extend(inherited_fields)
            setattr(obj_type, _CONFIG_FIELDS_KEY, config_fields)
        config_fields.append(self._to_kv_field(name))
        # Replace the UI config field spec with a regular dataclass default value
        setattr(obj_type, name, self.default)

    def _to_kv_field(self, name: str) -> SerializedKVConfigSchematicsField:
        raise NotImplementedError


@dataclass(frozen=True, slots=True)
class _ConfigBool(_ConfigField[bool]):
    """Boolean config field."""

    default: bool

    def _to_kv_field(self, name: str) -> SerializedKVConfigSchematicsField:
        return SerializedKVConfigSchematicsField(
            short_key=name,
            full_key=name,
            type_key="boolean",
            type_params={
                "displayName": self.label,
                "hint": self.hint,
            },
            default_value=self.default,
        )


@dataclass(frozen=True, slots=True)
class _ConfigInt(_ConfigField[int]):
    """Integer config field."""

    default: int

    def _to_kv_field(self, name: str) -> SerializedKVConfigSchematicsField:
        return SerializedKVConfigSchematicsField(
            short_key=name,
            full_key=name,
            type_key="numeric",
            type_params={
                "displayName": self.label,
                "hint": self.hint,
                "int": True,
            },
            default_value=self.default,
        )


@dataclass(frozen=True, slots=True)
class _ConfigFloat(_ConfigField[float]):
    """Floating point config field."""

    default: float

    def _to_kv_field(self, name: str) -> SerializedKVConfigSchematicsField:
        return SerializedKVConfigSchematicsField(
            short_key=name,
            full_key=name,
            type_key="numeric",
            type_params={
                "displayName": self.label,
                "hint": self.hint,
                "int": False,
            },
            default_value=self.default,
        )


@dataclass(frozen=True, slots=True)
class _ConfigString(_ConfigField[str]):
    """String config field."""

    default: str

    def _to_kv_field(self, name: str) -> SerializedKVConfigSchematicsField:
        return SerializedKVConfigSchematicsField(
            short_key=name,
            full_key=name,
            type_key="string",
            type_params={
                "displayName": self.label,
                "hint": self.hint,
            },
            default_value=self.default,
        )


@sdk_public_api()
def config_field(*, label: str, hint: str, default: _T) -> _T:
    """Define a plugin config field to be displayed and updated via the app UI."""
    # This type hint intentionally doesn't match the actual returned type
    # (the relevant ConfigField[_T] subclass). This is to ensure that
    # type checkers will accept config field initialisations like
    # "attr: int = config_field(...)".
    descriptor: _ConfigField[Any]
    match default:
        case bool():
            descriptor = _ConfigBool(label, hint, default)
        case float():
            descriptor = _ConfigFloat(label, hint, default)
        case int():
            descriptor = _ConfigInt(label, hint, default)
        case str():
            descriptor = _ConfigString(label, hint, default)
        case _:
            msg = f"Unsupported type for plugin config field: {type(default)!r}"
            raise LMStudioPluginInitError(msg)
    return cast(_T, descriptor)


# TODO: Cover additional config field types
# TODO: Allow optional constraints and UI display features
#       (the similarity of the _to_kv_field methods will reduce when this is done)


class _ImplicitDataClass(type):
    def __new__(
        meta_cls, name: str, bases: tuple[type, ...], namespace: dict[str, Any]
    ) -> type:
        cls: type = super().__new__(meta_cls, name, bases, namespace)
        return dataclass()(cls)


@dataclass_transform(field_specifiers=(config_field,))
@plugin_sdk_type
class BaseConfigSchema(metaclass=_ImplicitDataClass):
    """Base class for plugin configuration schema definitions."""

    # This uses the custom metaclass to automatically make subclasses data classes
    # Declare that behaviour in a way that mypy will fully accept
    __dataclass_fields__: ClassVar[dict[str, Field[Any]]]

    # ConfigField.__set_name__ lazily creates this class variable
    __kv_config_fields__: ClassVar[Sequence[SerializedKVConfigSchematicsField]]

    @classmethod
    def _to_kv_config_schematics(cls) -> SerializedKVConfigSchematics | None:
        """Convert to wire format for transmission to the app server."""
        try:
            config_fields = cls.__kv_config_fields__
        except AttributeError:
            # No config fields have been defined on this config schema
            # This is fine (as it allows for placeholders in code skeletons)
            return None
        return SerializedKVConfigSchematics(fields=config_fields)

    @classmethod
    def _default_config(cls) -> dict[str, Any]:
        default_config: dict[str, Any] = {}
        config_spec = cls()
        for field in fields(config_spec):
            attr = field.name
            default_config[attr] = getattr(config_spec, attr)
        return default_config

    @classmethod
    def _parse(cls, dynamic_config: SerializedKVConfigSettings) -> Self:
        config = cls._default_config()
        config.update(dict_from_kvconfig(dynamic_config))
        return cls(**config)



"""Test specific aspects of error and event logging."""

import logging

import anyio
import pytest
from pytest import LogCaptureFixture as LogCap

from lmstudio import AsyncClient

from .support import InvalidEndpoint


@pytest.mark.asyncio
@pytest.mark.slow
@pytest.mark.lmstudio
async def test_invalid_endpoint_request_stream(caplog: LogCap) -> None:
    caplog.set_level(logging.ERROR)
    async with AsyncClient() as client:
        session = client.llm
        # This will time out due to the bad API endpoint
        with pytest.raises(TimeoutError):
            with anyio.fail_after(1):
                endpoint = InvalidEndpoint()
                async with session._create_channel(endpoint) as channel:
                    async for _ in channel.rx_stream():
                        break
    logged_errors = caplog.records
    logged_error = logged_errors[0]
    logged_msg = logged_error.getMessage()
    assert "SDK communication warning" in logged_msg
    assert "noSuchEndpoint" in logged_msg
    assert logged_error.levelname.lower() == "error"
    assert len(logged_errors) == 1



"""Support for implementing LM Studio plugins in Python."""

# Using wildcard imports to export API symbols is acceptable
# ruff: noqa: F403

from .sdk_api import *
from .config_schemas import *
from .hooks import *
from .runner import *



"""Tests for default parameter values in tool definitions."""

import pytest

from msgspec import defstruct

from lmstudio.json_api import _NO_DEFAULT, ToolFunctionDef, ToolFunctionDefDict
from lmstudio.schemas import _to_json_schema


def greet(name: str, greeting: str = "Hello", punctuation: str = "!") -> str:
    """Greet someone with a customizable message.

    Args:
        name: The name of the person to greet
        greeting: The greeting word to use (default: "Hello")
        punctuation: The punctuation to end with (default: "!")

    Returns:
        A greeting message
    """
    return f"{greeting}, {name}{punctuation}"


def calculate(expression: str, precision: int = 2) -> str:
    """Calculate a mathematical expression.

    Args:
        expression: The mathematical expression to evaluate
        precision: Number of decimal places (default: 2)

    Returns:
        The calculated result as a string
    """
    return f"Result: {eval(expression):.{precision}f}"


class TestDefaultValues:
    """Test cases for default parameter values in tool definitions."""

    def test_extract_defaults_from_callable(self) -> None:
        """Test extracting default values from a callable function."""
        tool_def = ToolFunctionDef.from_callable(greet)

        assert tool_def.name == "greet"
        # Check that defaults are converted to inline format
        assert tool_def.parameters["greeting"] == {"type": str, "default": "Hello"}
        assert tool_def.parameters["punctuation"] == {"type": str, "default": "!"}
        assert tool_def.parameters["name"] is str  # No default, just type

    def test_manual_inline_defaults(self) -> None:
        """Test manually specifying default values in inline format."""
        tool_def = ToolFunctionDef(
            name="calculate",
            description="Calculate a mathematical expression",
            parameters={"expression": str, "precision": {"type": int, "default": 2}},
            implementation=calculate,
        )

        # Check that the inline format is preserved
        assert tool_def.parameters["precision"] == {"type": int, "default": 2}
        assert tool_def.parameters["expression"] is str  # No default, just type

    def test_json_schema_with_defaults(self) -> None:
        """Test that JSON schema includes default values."""
        tool_def = ToolFunctionDef.from_callable(greet)
        params_struct, _ = tool_def._to_llm_tool_def()

        json_schema = _to_json_schema(params_struct)

        # Check that default values are included in the schema
        assert json_schema["properties"]["greeting"]["default"] == "Hello"
        assert json_schema["properties"]["punctuation"]["default"] == "!"
        assert "default" not in json_schema["properties"]["name"]

    def test_dict_based_definition(self) -> None:
        """Test dictionary-based tool definition with inline defaults."""
        dict_tool: ToolFunctionDefDict = {
            "name": "format_text",
            "description": "Format text with specified style",
            "parameters": {
                "text": str,
                "style": {"type": str, "default": "normal"},
                "uppercase": {"type": bool, "default": False},
            },
            "implementation": lambda text, style="normal", uppercase=False: text.upper()
            if uppercase
            else text,
        }

        # This should work without errors
        tool_def = ToolFunctionDef(**dict_tool)
        assert tool_def.parameters["style"] == {"type": str, "default": "normal"}
        assert tool_def.parameters["uppercase"] == {"type": bool, "default": False}
        assert tool_def.parameters["text"] is str  # No default, just type

    def test_no_defaults(self) -> None:
        """Test function with no default values."""

        def no_defaults(a: int, b: str) -> str:
            """Function with no default parameters."""
            return f"{a}: {b}"

        tool_def = ToolFunctionDef.from_callable(no_defaults)
        # All parameters should be simple types without defaults
        assert tool_def.parameters["a"] is int
        assert tool_def.parameters["b"] is str

        params_struct, _ = tool_def._to_llm_tool_def()
        json_schema = _to_json_schema(params_struct)

        # No default values should be present
        assert "default" not in json_schema["properties"]["a"]
        assert "default" not in json_schema["properties"]["b"]

    def test_mixed_defaults(self) -> None:
        """Test function with some parameters having defaults and others not."""

        def mixed_defaults(
            required: str, optional1: int = 42, optional2: bool = True
        ) -> str:
            """Function with mixed required and optional parameters."""
            return f"{required}: {optional1}, {optional2}"

        tool_def = ToolFunctionDef.from_callable(mixed_defaults)
        # Check inline format for parameters with defaults
        assert tool_def.parameters["optional1"] == {"type": int, "default": 42}
        assert tool_def.parameters["optional2"] == {"type": bool, "default": True}
        assert tool_def.parameters["required"] is str  # No default, just type

        params_struct, _ = tool_def._to_llm_tool_def()
        json_schema = _to_json_schema(params_struct)

        # Check that default values are correctly included in schema
        assert json_schema["properties"]["optional1"]["default"] == 42
        assert json_schema["properties"]["optional2"]["default"] is True
        assert "default" not in json_schema["properties"]["required"]

    def test_extract_type_and_default_method(self) -> None:
        """Test the _extract_type_and_default helper method."""

        # Test simple type
        param_type, default = ToolFunctionDef._extract_type_and_default(str)
        assert param_type is str
        assert default is _NO_DEFAULT

        # Test inline format with missing type key
        with pytest.raises(TypeError, match="Missing 'type' key"):
            param_type, default = ToolFunctionDef._extract_type_and_default(
                {"default": 42}  # type: ignore[arg-type]
            )

        # Test inline format with no default
        param_type, default = ToolFunctionDef._extract_type_and_default({"type": int})
        assert param_type is int
        assert default is _NO_DEFAULT

        # Test inline format with default
        param_type, default = ToolFunctionDef._extract_type_and_default(
            {"type": int, "default": 42}
        )
        assert param_type is int
        assert default == 42

        # Test complex default
        param_type, default = ToolFunctionDef._extract_type_and_default(
            {"type": list, "default": [1, 2, 3]}
        )
        assert param_type is list
        assert default == [1, 2, 3]

    def test_msgspec_auto_defaults(self) -> None:
        """msgspec automatically reflects default values in the JSON schema."""
        TestStruct = defstruct(
            "TestStruct",
            [
                ("name", str),
                ("age", int, 18),
                ("active", bool, True),
            ],
            kw_only=True,
        )

        schema = _to_json_schema(TestStruct)
        properties = schema.get("properties", {})
        required = schema.get("required", [])

        assert "name" in properties and "default" not in properties["name"]
        assert properties["age"].get("default") == 18
        assert properties["active"].get("default") is True
        assert "name" in required and "age" not in required and "active" not in required



"""Plugin dev client implementation."""

import asyncio
import io
import os
import signal
import subprocess
import sys

from contextlib import asynccontextmanager
from functools import partial
from pathlib import Path
from threading import Event as SyncEvent
from typing import Any, AsyncGenerator, Iterable, TypeAlias

from typing_extensions import (
    # Native in 3.11+
    assert_never,
)

from .runner import (
    ENV_CLIENT_ID,
    ENV_CLIENT_KEY,
    PluginClient,
)
from ..schemas import DictObject
from ..json_api import (
    ChannelCommonRxEvent,
    ChannelEndpoint,
    ChannelFinishedEvent,
    ChannelRxEvent,
    LMStudioChannelClosedError,
)
from .._sdk_models import (
    # TODO: Define aliases at schema generation time
    PluginsChannelRegisterDevelopmentPluginCreationParameter as DevPluginRegistrationRequest,
    PluginsChannelRegisterDevelopmentPluginCreationParameterDict as DevPluginRegistrationRequestDict,
    PluginsChannelRegisterDevelopmentPluginToServerPacketEndDict as DevPluginRegistrationEndDict,
)


class DevPluginRegistrationReadyEvent(ChannelRxEvent[None]):
    pass


DevPluginRegistrationRxEvent: TypeAlias = (
    DevPluginRegistrationReadyEvent | ChannelCommonRxEvent
)


class DevPluginRegistrationEndpoint(
    ChannelEndpoint[
        tuple[str, str], DevPluginRegistrationRxEvent, DevPluginRegistrationRequestDict
    ]
):
    """API channel endpoint to register a development plugin and receive credentials."""

    _API_ENDPOINT = "registerDevelopmentPlugin"
    _NOTICE_PREFIX = "Register development plugin"

    def __init__(self, owner: str, name: str) -> None:
        # TODO: Set "python" as the type once LM Studio supports that
        params = DevPluginRegistrationRequest._from_api_dict(
            {
                "manifest": {
                    "type": "plugin",
                    "runner": "node",
                    "owner": owner,
                    "name": name,
                }
            }
        )
        super().__init__(params)

    def iter_message_events(
        self, contents: DictObject | None
    ) -> Iterable[DevPluginRegistrationRxEvent]:
        match contents:
            case None:
                raise LMStudioChannelClosedError(
                    "Server failed to complete development plugin registration."
                )
            case {
                "type": "ready",
                "clientIdentifier": str(client_id),
                "clientPasskey": str(client_key),
            }:
                yield self._set_result((client_id, client_key))
            case unmatched:
                self.report_unknown_message(unmatched)

    def handle_rx_event(self, event: DevPluginRegistrationRxEvent) -> None:
        match event:
            case DevPluginRegistrationReadyEvent(_):
                pass
            case ChannelFinishedEvent(_):
                pass
            case _:
                assert_never(event)


class DevPluginClient(PluginClient):
    def _get_registration_endpoint(self) -> DevPluginRegistrationEndpoint:
        return DevPluginRegistrationEndpoint(self.owner, self.name)

    @asynccontextmanager
    async def register_dev_plugin(self) -> AsyncGenerator[tuple[str, str], None]:
        """Register a dev plugin on entry, deregister it on exit."""
        endpoint = self._get_registration_endpoint()
        async with self.plugins._create_channel(endpoint) as channel:
            registration_result = await channel.wait_for_result()
            try:
                yield registration_result
            finally:
                message: DevPluginRegistrationEndDict = {"type": "end"}
                await channel.send_message(message)

    async def _run_plugin_task(
        self, result_queue: asyncio.Queue[int], debug: bool = False
    ) -> None:
        notify_subprocess_thread = SyncEvent()
        async with self.register_dev_plugin() as (client_id, client_key):
            wait_for_subprocess = asyncio.ensure_future(
                asyncio.to_thread(
                    partial(
                        _run_plugin_in_child_process,
                        self._plugin_path,
                        client_id,
                        client_key,
                        notify_subprocess_thread,
                        debug=debug,
                    )
                )
            )
            try:
                result = await wait_for_subprocess
            except asyncio.CancelledError:
                # Likely a Ctrl-C press, which is the expected termination process
                notify_subprocess_thread.set()
                result_queue.put_nowait(0)
                raise
            # Subprocess terminated, pass along its return code in the parent process
            await result_queue.put(result)

    async def run_plugin(
        self, *, allow_local_imports: bool = True, debug: bool = False
    ) -> int:
        if not allow_local_imports:
            raise ValueError("Local imports are always permitted for dev plugins")
        result_queue: asyncio.Queue[int] = asyncio.Queue()
        # Run in the task manager, so this gets cleaned up before the websocket handler
        await self._task_manager.schedule_task(
            partial(self._run_plugin_task, result_queue, debug)
        )
        return await result_queue.get()


def _get_creation_flags() -> int:
    if sys.platform == "win32":
        return subprocess.CREATE_NEW_PROCESS_GROUP
    return 0


def _start_child_process(
    command: list[str], *, text: bool | None = True, **kwds: Any
) -> subprocess.Popen[str]:
    creationflags = kwds.pop("creationflags", 0)
    creationflags |= _get_creation_flags()
    return subprocess.Popen(command, text=text, creationflags=creationflags, **kwds)


def _get_interrupt_signal() -> signal.Signals:
    if sys.platform == "win32":
        return signal.CTRL_C_EVENT
    return signal.SIGINT


_PLUGIN_INTERRUPT_SIGNAL = _get_interrupt_signal()
_PLUGIN_STATUS_POLL_INTERVAL = 1
_PLUGIN_STOP_TIMEOUT = 2


def _interrupt_child_process(process: subprocess.Popen[Any], timeout: float) -> int:
    process.send_signal(_PLUGIN_INTERRUPT_SIGNAL)
    try:
        return process.wait(timeout)
    except TimeoutError:
        process.kill()
        raise


# TODO: support the same source code change monitoring features as `lms dev`
def _run_plugin_in_child_process(
    plugin_path: Path,
    client_id: str,
    client_key: str,
    abort_event: SyncEvent,
    *,
    debug: bool = False,
) -> int:
    env = os.environ.copy()
    env[ENV_CLIENT_ID] = client_id
    env[ENV_CLIENT_KEY] = client_key
    package_name = __spec__.parent
    assert package_name is not None
    debug_option = ("--debug",) if debug else ()
    # If stdout is unbuffered, specify the same in the child process
    stdout = sys.__stdout__
    unbuffered_arg: tuple[str, ...]
    if stdout is None or not isinstance(stdout.buffer, io.BufferedWriter):
        unbuffered_arg = ("-u",)
    else:
        unbuffered_arg = ()

    command: list[str] = [
        sys.executable,
        *unbuffered_arg,
        "-m",
        package_name,
        *debug_option,
        os.fspath(plugin_path),
    ]
    process = _start_child_process(command, env=env)
    while True:
        result = process.poll()
        if result is not None:
            print("Child process terminated unexpectedly")
            break
        if abort_event.wait(_PLUGIN_STATUS_POLL_INTERVAL):
            print("Gracefully terminating child process...")
            result = _interrupt_child_process(process, _PLUGIN_STOP_TIMEOUT)
            break
    return result


async def run_plugin_async(
    plugin_dir: str | os.PathLike[str], *, debug: bool = False
) -> int:
    """Asynchronously execute a plugin in development mode."""
    async with DevPluginClient(plugin_dir) as dev_client:
        return await dev_client.run_plugin(debug=debug)


def run_plugin(plugin_dir: str | os.PathLike[str], *, debug: bool = False) -> int:
    """Execute a plugin in development mode."""
    return asyncio.run(run_plugin_async(plugin_dir, debug=debug))
